# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module implements the Electrode class used for calculating
relevant electrode properties from phase diagrams.

"""

__all__ = ['Electrode', 'VoltageProfile']
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from .electrode import Electrode, VoltageProfile



#!/usr/bin/env python
import unittest


class ConnectTest(unittest.TestCase):
    """ Test simple database connection methods. """

    def testFuzzyCollnames(self):
        """ Test the fuzzy collection name matcher. """
        from matador.db.connect import fuzzy_collname_match

        trial = "me388_proto_test"
        targets = [
            "me388_prototype_test",
            "proto_test_me388",
            "me388_fun",
            "ajm255_test",
            "ajm255_no",
        ]
        options = fuzzy_collname_match(trial, targets)

        results = [targets[0], targets[1], targets[2], targets[3]]
        self.assertEqual(results, options)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Distributed under the terms of the MIT license.

""" This file implements the base DataContainer class which wraps raw
matador dictionaries and adds useful methods to be inherited by its
children.

"""

import copy
import math
from abc import ABC


class DataContainer(ABC):
    """ Base class for matador data classes. This class is a read-only
    store of the underlying dictionary of raw data; its children can
    implement useful methods to inspect and analyse the underlying data.

    """

    required_keys = []

    def __init__(self, data=None, **kwargs):
        """ Initalise copy of raw data. """
        if isinstance(data, dict):
            self._data = copy.deepcopy(data)
        else:
            self._data = {key: kwargs[key] for key in kwargs}

        self._validate_inputs()

        # set root source to source filename
        from matador.utils.chem_utils import get_root_source
        self.root_source = 'unknown'
        try:
            if 'source' in self._data:
                self.root_source = get_root_source(self._data['source'])
        except RuntimeError:
            pass

    @property
    def source(self):
        """ Return the source of the data. """
        if 'source' not in self._data:
            return 'unknown'
        return self._data['source']

    def _validate_inputs(self):
        """ Validate the incoming data by checking the existence
        of the required keys for the subclass.

        """
        missing_keys = []
        for key in self.required_keys:
            if key not in self._data:
                missing_keys.append(key)

        if missing_keys:
            raise RuntimeError(
                "Unable to create object of type {} as the following keys are missing: {}"
                .format(type(self).__name__, missing_keys)
            )

    def __getitem__(self, key):
        """ Allow properties to be used with key access.

        Parameters:
            key (str): name of key or attribute to get.

        Raises:
            AttributeError: if key or attribute can't be found.

        """
        try:
            return getattr(self, key)
        except AttributeError:
            pass

        try:
            return self._data[key]
        except KeyError:
            raise KeyError('Object has no data or implementation for requested key: "{}"'
                           .format(key))

    def __delitem__(self, key: str):
        raise AttributeError('Object does not support deletion of keys in `_data`.')

    def __setitem__(self, key: str, item):
        if key not in self._data or self._data[key] is None:
            self._data[key] = item
            return
        elif self._data[key] != item:
            try:
                if (math.isnan(item) and math.isnan(self._data[key])):
                    return
            except TypeError:
                pass
            raise AttributeError('Cannot assign value {} to existing key {} with value {}'
                                 .format(item, key, self._data[key]))

    def __contains__(self, key):
        if key in self._data:
            return True
        return False

    def get(self, *args):
        """ Overload dictionary.get() method.

        Parameters:
            key (str): key to try and obtain.

        Keyword arguments:
            default: return value raise if key is not present.

        Returns:
            :attr:`_data[key]` if it exists, else None.

        """
        key = args[0]
        if len(args) > 1:
            default = args[1]
        else:
            default = None
        if len(args) > 2:
            raise TypeError('get() takes up to 2 arguments, not {}'.format(len(args)))

        return self._data.get(key, default)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module contains various utility functions that
are used liberally throughout matador.

"""


__all__ = []
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'



#!/usr/bin/env python

import unittest
import os
import numpy as np
from matador.utils.chem_utils import get_concentration
from matador.utils.chem_utils import (
    get_generic_grav_capacity,
    get_binary_volumetric_capacity,
)
from matador.utils.chem_utils import (
    get_stoich,
    get_formula_from_stoich,
    get_stoich_from_formula,
)
from matador.utils.chem_utils import get_ratios_from_stoichiometry, get_root_source
from matador.utils.chem_utils import get_formation_energy
from matador.utils.chem_utils import get_padded_composition

REAL_PATH = "/".join(os.path.realpath(__file__).split("/")[:-1]) + "/"


class ChemUtilsTest(unittest.TestCase):
    """ Test chem utils functionality. """

    def test_gravimetric_capacity(self):
        test_docs = []
        test_elements = []
        Q = []

        # Li3P and symmetry test
        doc = dict()
        doc["stoichiometry"] = [["Li", 3], ["P", 1]]
        test_docs.append(doc)
        test_elements.append(["Li", "P"])
        doc = dict()
        doc["stoichiometry"] = [["P", 1], ["Li", 3]]
        test_docs.append(doc)
        test_elements.append(["Li", "P"])
        # ternary test
        doc = dict()
        doc["stoichiometry"] = [["Li", 1], ["Mo", 1], ["S", 2]]
        test_docs.append(doc)
        test_elements.append(["Li", "Mo", "S"])
        doc = dict()
        doc["stoichiometry"] = [["Li", 8], ["Mo", 1], ["S", 2]]
        test_docs.append(doc)
        test_elements.append(["Li", "Mo", "S"])
        doc = dict()
        doc["stoichiometry"] = [["Li", 1], ["Mo", 2], ["S", 4]]
        test_docs.append(doc)
        test_elements.append(["Li", "Mo", "S"])
        for doc, elem in zip(test_docs, test_elements):
            doc["concentration"] = get_concentration(doc, elem)
            temp_conc = list(doc["concentration"])
            temp_conc.append(1.0)
            for conc in doc["concentration"]:
                temp_conc[-1] -= conc

            Q.append(get_generic_grav_capacity(temp_conc, elem))
        self.assertAlmostEqual(Q[0], 2596.096626125, places=3)
        self.assertAlmostEqual(Q[2], 167.449, places=3)
        self.assertEqual(Q[0], Q[1])
        self.assertEqual(round(8 * Q[2], 3), round(Q[3], 3))
        self.assertEqual(round(Q[2], 3), round(2 * Q[4], 3))

    def test_get_concentration(self):
        doc = [["Na", 3], ["P", 1]]
        conc = get_concentration(doc, ["Na", "P"])
        self.assertEqual(conc, [0.75])
        conc = get_concentration(doc, ["Na", "P"], include_end=True)
        self.assertEqual(conc, [0.75, 0.25])
        conc = get_concentration(doc, ["K", "P"], include_end=True)
        self.assertEqual(conc, [0.0, 0.25])

        doc = {"stoichiometry": [["Na", 3], ["P", 1]]}
        conc = get_concentration(doc, ["Na", "P"])
        self.assertEqual(conc, [0.75])
        conc = get_concentration(doc, ["Na", "P"], include_end=True)
        self.assertEqual(conc, [0.75, 0.25])

    def test_volumetric_capacity(self):
        initial_doc = dict()
        final_doc = dict()
        initial_doc["stoichiometry"] = [["P", 1]]
        initial_doc["cell_volume"] = 84.965349
        initial_doc["num_fu"] = 4

        final_doc["stoichiometry"] = sorted([["Li", 3], ["P", 1]])
        vol_cap = get_binary_volumetric_capacity(initial_doc, final_doc)
        self.assertAlmostEqual(vol_cap, 6286, places=0)

    def test_atoms_to_stoich(self):
        atoms = 5 * ["Li"]
        atoms.extend(5 * ["P"])
        stoich = [["Li", 1], ["P", 1]]
        self.assertEqual(stoich, get_stoich(atoms))
        atoms = 99 * ["Li"]
        atoms.extend(1 * ["P"])
        stoich = [["Li", 99], ["P", 1]]
        self.assertEqual(stoich, get_stoich(atoms))
        atoms = 4 * ["Li"]
        atoms.extend(36 * ["P"])
        stoich = [["Li", 1], ["P", 9]]
        self.assertEqual(stoich, get_stoich(atoms))
        atoms = 3 * ["Li"]
        atoms.extend(2 * ["P"])
        stoich = [["Li", 3], ["P", 2]]
        self.assertEqual(stoich, get_stoich(atoms))
        atoms = 9 * ["Li"]
        atoms.extend(6 * ["P"])
        stoich = [["Li", 3], ["P", 2]]
        self.assertEqual(stoich, get_stoich(atoms))
        atoms = 36 * ["P"]
        atoms.extend(4 * ["Li"])
        stoich = [["Li", 1], ["P", 9]]
        self.assertEqual(stoich, get_stoich(atoms))

    def test_stoich_to_form(self):
        stoich = [["Li", 1], ["P", 9]]
        form = "LiP9"
        self.assertEqual(form, get_formula_from_stoich(stoich))
        stoich = [["P", 9], ["Li", 1]]
        form = "LiP9"
        self.assertEqual(form, get_formula_from_stoich(stoich))
        stoich = [["Li", 1], ["P", 9]]
        form = "LiP$_\\mathrm{9}$"
        self.assertEqual(
            form, get_formula_from_stoich(stoich, tex=True, latex_sub_style="\\mathrm")
        )
        stoich = [["Li", 1], ["P", 9]]
        form = "P9Li"
        self.assertEqual(form, get_formula_from_stoich(stoich, elements=["P", "Li"]))

    def test_form_to_stoich(self):
        formula = "Li12P1N18"
        stoich = [["Li", 12], ["P", 1], ["N", 18]]
        self.assertEqual(stoich, get_stoich_from_formula(formula, sort=False))

    def test_padded_composition(self):
        stoich = [["Li", 2], ["O", 1]]
        elements = ["O", "Li", "Ba"]
        self.assertEqual(get_padded_composition(stoich, elements), [1, 2, 0])
        elements = ["O", "Ba", "Li"]
        self.assertEqual(get_padded_composition(stoich, elements), [1, 0, 2])
        elements = ["Ba", "Li", "O"]
        self.assertEqual(get_padded_composition(stoich, elements), [0, 2, 1])
        stoich = [["F", 8]]
        elements = ["F"]
        self.assertEqual(get_padded_composition(stoich, elements), [8])
        elements = ["Fe", "Ba", "Ca"]
        with self.assertRaises(RuntimeError):
            get_padded_composition(stoich, elements)

    def test_formation_energy_binary(self):
        chempots = []
        chempots.append(
            {
                "stoichiometry": [["Li", 1]],
                "enthalpy": -200,
                "enthalpy_per_atom": -100,
                "temperature": {"free_energy_per_atom": {"300": -100, "400": -120}},
                "num_fu": 2,
                "atom_types": ["Li", "Li"],
            }
        )
        chempots.append(
            {
                "stoichiometry": [["P", 1]],
                "enthalpy": -600,
                "enthalpy_per_atom": -200,
                "num_fu": 3,
                "num_atoms": 3,
                "temperature": {
                    "free_energy_per_atom": {"300": -200, "400": -210, "450": -300}
                },
                "atom_types": ["P", "P", "P"],
            }
        )
        cursor = []
        cursor.append(
            {
                "stoichiometry": [["Li", 1], ["P", 1]],
                "enthalpy": -1200,
                "enthalpy_per_atom": -300,
                "num_fu": 2,
                "num_atoms": 4,
                "temperature": {"free_energy_per_atom": {"300": -300, "400": -315}},
                "atom_types": ["Li", "Li", "P", "P"],
            }
        )

        cursor.append(
            {
                "stoichiometry": [["Li", 3], ["P", 1]],
                "enthalpy": -1600,
                "enthalpy_per_atom": -400,
                "num_fu": 1,
                "num_atoms": 4,
                "atom_types": ["Li", "Li", "Li", "P"],
            }
        )

        ef = []
        for doc in cursor:
            ef.append(get_formation_energy(chempots, doc))

        self.assertEqual(ef[0], -600 / 4)
        self.assertEqual(ef[1], -1100 / 4)

        ef_300 = get_formation_energy(
            chempots,
            cursor[0],
            energy_key=["temperature", "free_energy_per_atom", "300"],
        )
        self.assertEqual(ef_300, -150)
        ef_400 = get_formation_energy(
            chempots,
            cursor[0],
            energy_key=["temperature", "free_energy_per_atom", "400"],
        )
        self.assertEqual(ef_400, -150)

        with self.assertRaises(KeyError):
            get_formation_energy(
                chempots,
                cursor[0],
                energy_key=["temperature", "free_energy_per_atom", "450"],
            )

    def test_formation_energy_nonbinary(self):
        from matador.utils.chem_utils import get_formation_energy

        chempots = []
        chempots.append(
            {
                "stoichiometry": [["Li", 2], ["O", 1]],
                "enthalpy": -1200,
                "enthalpy_per_atom": -200,
                "num_fu": 2,
                "num_atoms": 6,
                "atom_types": ["Li", "Li", "Li", "Li", "O", "O"],
            }
        )
        chempots.append(
            {
                "stoichiometry": [["La", 2], ["O", 3]],
                "enthalpy": -500,
                "enthalpy_per_atom": -100,
                "num_fu": 1,
                "num_atoms": 5,
                "atom_types": ["La", "La", "O", "O", "O"],
            }
        )
        chempots.append(
            {
                "stoichiometry": [["Zr", 1], ["O", 2]],
                "enthalpy": -4500,
                "enthalpy_per_atom": -150,
                "num_fu": 10,
                "num_atoms": 30,
            }
        )
        cursor = []
        cursor.append(
            {
                "stoichiometry": [["La", 2], ["Li", 4], ["O", 5]],
                "enthalpy": -1700,
                "enthalpy_per_atom": -1700 / 11,
                "num_fu": 1,
                "num_atoms": 11,
                "atom_types": [
                    "La",
                    "La",
                    "Li",
                    "Li",
                    "Li",
                    "Li",
                    "O",
                    "O",
                    "O",
                    "O",
                    "O",
                ],
            }
        )
        cursor.append(
            {
                "stoichiometry": [["La", 2], ["Li", 4], ["O", 5]],
                "enthalpy": -1711,
                "enthalpy_per_atom": -1711 / 11,
                "num_fu": 1,
                "num_atoms": 11,
                "atom_types": [
                    "La",
                    "La",
                    "Li",
                    "Li",
                    "Li",
                    "Li",
                    "O",
                    "O",
                    "O",
                    "O",
                    "O",
                ],
            }
        )
        cursor.append(
            {
                "stoichiometry": [["Li", 7], ["La", 3], ["Zr", 2], ["O", 12]],
                "enthalpy": -3.5 * 3 * 200 - 1.5 * 5 * 100 - 2 * 3 * -150,
                "enthalpy_per_atom": (-3.5 * 3 * 200 - 1.5 * 5 * 100 - 2 * 3 * 150)
                / 24,
            }
        )
        cursor.append(
            {
                "stoichiometry": [["Li", 7], ["La", 3], ["Zr", 2], ["O", 12]],
                "enthalpy": -3.5 * 3 * 200 - 1.5 * 5 * 100 - 2 * 3 * 150 - 48,
                "enthalpy_per_atom": (-3.5 * 3 * 200 - 1.5 * 5 * 100 - 2 * 3 * 150 - 48)
                / 24,
            }
        )

        ef = []
        for doc in cursor:
            ef.append(get_formation_energy(chempots, doc))

        np.testing.assert_array_almost_equal(ef, [0, -1, 0, -2])

    def test_get_number_of_chempots(self):
        from matador.utils.chem_utils import get_number_of_chempots

        stoich = [["V", 2]]
        chempots = [[["Na", 1]], [["S", 1], ["O", 4]], [["V", 1]]]
        self.assertEqual(get_number_of_chempots(stoich, chempots), [0, 0, 2])

        stoich = [["Li", 7], ["La", 3], ["Zr", 2], ["O", 12]]
        chempots = [[["Li", 2], ["O", 1]], [["Zr", 1], ["O", 2]], [["La", 2], ["O", 3]]]
        self.assertEqual(get_number_of_chempots(stoich, chempots), [3.5, 2, 1.5])

        stoich = [["Li", 7], ["La", 3], ["Zr", 2], ["O", 12]]
        chempots = [[["Li", 2], ["O", 1]], [["Zr", 1], ["O", 2]], [["La", 2], ["O", 3]]]
        np.testing.assert_array_almost_equal(
            get_number_of_chempots(stoich, chempots, precision=None), [3.5, 2, 1.5]
        )

        stoich = [["Li", 8], ["Zr", 3], ["O", 10]]
        chempots = [[["Li", 2], ["O", 1]], [["Zr", 1], ["O", 2]], [["La", 2], ["O", 3]]]
        self.assertEqual(get_number_of_chempots(stoich, chempots), [4, 3, 0])

        stoich = [["Li", 3], ["P", 1]]
        chempots = [[["Li", 1]], [["Zn", 1]], [["P", 1]]]
        self.assertEqual(get_number_of_chempots(stoich, chempots), [3, 0, 1])

        stoich = [["Li", 1]]
        chempots = [[["Li", 1]], [["Zn", 1]], [["P", 1]]]
        self.assertEqual(get_number_of_chempots(stoich, chempots), [1, 0, 0])

        stoich = [["La", 2], ["Li", 4], ["O", 5]]
        chempots = [[["La", 2], ["O", 3]], [["Li", 2], ["O", 1]]]
        self.assertEqual(get_number_of_chempots(stoich, chempots), [1, 2])

        stoich = [["Li", 1], ["P", 1], ["O", 3]]
        chempots = [[["Li", 2], ["S", 1]], [["P", 2], ["O", 5]], [["P", 2], ["S", 5]]]
        with self.assertRaises(RuntimeError):
            get_number_of_chempots(stoich, chempots)

        stoich = [["Na", 1], ["B", 1], ["S", 1], ["O", 4]]
        chempots = [[["Na", 1]], [["B", 1]], [["S", 1], ["O", 4]]]
        self.assertEqual(get_number_of_chempots(stoich, chempots), [1, 1, 1])

        stoich = [["Na", 1]]
        chempots = [[["Na", 1]], [["B", 1]], [["S", 1], ["O", 4]]]
        self.assertEqual(get_number_of_chempots(stoich, chempots), [1, 0, 0])

        stoich = [["B", 1]]
        chempots = [[["Na", 1]], [["B", 1]], [["S", 1], ["O", 4]]]
        self.assertEqual(get_number_of_chempots(stoich, chempots), [0, 1, 0])

        stoich = [["S", 2], ["O", 8]]
        chempots = [[["Na", 1]], [["B", 1]], [["S", 1], ["O", 4]]]
        self.assertEqual(get_number_of_chempots(stoich, chempots), [0, 0, 2])

        stoich = [["B", 2], ["Na", 9]]
        chempots = [[["Na", 1]], [["S", 1], ["O", 4]], [["B", 1]]]
        self.assertEqual(get_number_of_chempots(stoich, chempots), [9, 0, 2])

        stoich = [["B", 2], ["Na", 9]]
        chempots = [[["Na", 1]], [["S", 1], ["O", 4]], [["B", 2]]]
        self.assertEqual(get_number_of_chempots(stoich, chempots), [9, 0, 1])

        stoich = [["A", 1], ["B", 2]]
        chempots = [[["A", 1], ["B", 2], ["C", 3]]]
        with self.assertRaises(RuntimeError):
            get_number_of_chempots(stoich, chempots)

    def test_ratios_from_stoich(self):
        stoich = [["Li", 12], ["N", 18], ["P", 1]]
        ratios = {
            "LiN": round(12.0 / 18, 3),
            "LiP": 12,
            "NP": 18,
            "NLi": round(18.0 / 12, 3),
            "PLi": round(1.0 / 12, 3),
            "PN": round(1.0 / 18, 3),
        }
        self.assertEqual(ratios, get_ratios_from_stoichiometry(stoich))

        stoich = [["K", 8], ["Sn", 1], ["P", 4]]
        ratios = {
            "KSn": 8,
            "KP": 2,
            "SnP": 0.25,
            "SnK": round(1.0 / 8, 3),
            "PK": 0.5,
            "PSn": 4,
        }
        self.assertEqual(ratios, get_ratios_from_stoichiometry(stoich))

    def test_root_source(self):
        source = ["KP.cell", "KP.param", "KP.castep"]
        src = "KP"
        self.assertEqual(src, get_root_source(source))

        source = ["KP.cell", "KP.param", "KP-1234-abcd.castep"]
        src = "KP-1234-abcd"
        self.assertEqual(src, get_root_source(source))

        source = [
            "KP.cell",
            "KP.param",
            "abcd-123.fdasf/efgf/KP-0.02.-1234-abcd.castep",
            "KP-0.02.-1234-abcd.res",
        ]
        src = "KP-0.02.-1234-abcd"
        self.assertEqual(src, get_root_source(source))

        source = [
            "KP.cell",
            "KP.param",
            "abcd-123.fdasf/efgf/KP-0.02.-1234-abcd.history",
        ]
        src = "KP-0.02.-1234-abcd"
        self.assertEqual(src, get_root_source(source))

        source = ["KP.cell", "KP.param", "PK-OQMD_12345.history"]
        src = "PK-OQMD_12345"
        self.assertEqual(src, get_root_source(source))

        source = ["OQMD 12345"]
        src = "OQMD 12345"
        self.assertEqual(src, get_root_source(source))

        source = ["OQMD-12345"]
        src = "OQMD-12345"
        self.assertEqual(src, get_root_source(source))

        source = [
            "KP.cell",
            "KP.param",
            "abcd-123.fdasf/efgf/KP-0.02.-1234-abcd.castep",
            "KP-1234-abcde.res",
        ]
        with self.assertRaises(RuntimeError):
            src = get_root_source(source)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This module contains functionality for wrapping matador
data dictionaries as :obj:`Crystal` objects, which possess
several convenience methods.

"""


__all__ = ['Crystal']

__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'


from matador.crystal.crystal import Crystal



# encoding: utf-8
""" This file implements lattice-site level measures of similarity, using
Voronoi decompositions.

Sketch of methodology:

* For a list of structures, take each in turn, compute the Voronoi substructure
and create normalised padded arrays so that they can be compared.

* For each structure, find the *unique* sites, to some definition of unique, initially
just a simple np.isclose() on the site arrays.

* Now do an all-to-all comparison of the unique sites in each structure, yielding
an overall list of unique substructures. This step should have a dial that can be turned
such that all sites fold onto each other, or all sites become distinct, i.e. sensitivity
vs specificity.

* [OPTIONAL] The remaining unique sites across the whole list can now be clustered, if desired,
using e.g. hierarchical clustering.

* Every site in every structure can now be assigned to one of the unique sites above, or
one of the unique clusters.

* Finally, the structures themselves can be clustered by the sites that are present, if desired.

"""
from collections import defaultdict
import numpy as np

from matador.utils.cell_utils import frac2cart


def are_sites_the_same(site_A, site_B, rtol=1e-2, atol=1e-2):
    """ Simple check for uniqueness of sites based on their
    Voronoi substructure.

    Input:

        | site_A : dict, with elements as keys, containing numpy arrays
                   of normalised solid angles.
        | site_B : dict, as A.

    Args:

        | rtol : relative tolerance in solid angle for np.allclose,
        | atol : absolute tolerance in solid angle for np.allclose.

    Returns:

        | True if sites have approximately the same values for each species,
          else False.

    """
    # if a key is missing from A or B, check that its not some extraneous distant atom
    for key in site_B:
        if key not in site_A:
            site_A_missing_key = np.zeros_like(site_B[key])
            if not (np.allclose(site_A_missing_key, site_B[key], atol=atol, rtol=rtol) or np.allclose(site_B[key], site_A_missing_key, atol=atol, rtol=rtol)):
                return False
    # now check remaining keys
    for key in site_A:
        if key not in site_B:
            site_B_missing_key = np.zeros_like(site_A[key])
            if not (np.allclose(site_B_missing_key, site_A[key], atol=atol, rtol=rtol) or np.allclose(site_A[key], site_B_missing_key, atol=atol, rtol=rtol)):
                return False
        else:
            if len(site_A[key]) != len(site_B[key]):
                if len(site_A[key]) > len(site_B[key]):
                    padded_B = np.append(site_B[key], np.zeros((len(site_A[key])-len(site_B[key]))))
                    if not (np.allclose(site_A[key], padded_B, atol=atol, rtol=rtol) or np.allclose(site_A[key], padded_B, atol=atol, rtol=rtol)):
                        return False
                else:
                    padded_A = np.append(site_A[key], np.zeros((len(site_B[key])-len(site_A[key]))))
                    assert np.allclose(padded_A, site_B[key], atol=atol, rtol=rtol) is np.allclose(site_B[key], padded_A, atol=atol, rtol=rtol)
                    if not (np.allclose(padded_A, site_B[key], atol=atol, rtol=rtol) or np.allclose(site_B[key], padded_A, atol=atol, rtol=rtol)):
                        return False
            else:
                if not (np.allclose(site_A[key], site_B[key], atol=atol, rtol=rtol) or np.allclose(site_B[key], site_A[key], rtol=rtol, atol=atol)):
                    return False

    return True


def collect_unique_sites(cursor):
    """ Collect unique substrucs from each structure into a single dict.

    Input:

        | cursor: list(dict), list of structures with pre-computed unique sites.

    Returns:

        | unique_environments: dict(list), dict with element symbol keys containing
                               a list of each unique substructure.

    """
    unique_environments = defaultdict(list)
    for doc in cursor:
        for elem in doc['unique_substrucs']:
            for substruc in doc['unique_substrucs'][elem]:
                unique_environments[elem].append(substruc)
    return unique_environments


def get_max_coordination_of_elem(single_elem_environments):
    """ For a given set of elemental environments, find the greatest
    number of each element in a site.

    Input:

        | single_elem_environments: list(dict), list of voronoi substructures,

    Returns:

        | max_num_elem: dict(int), greatest number of each element in a substruc.

    """
    max_num_elem = {}
    for site_ind, site in enumerate(single_elem_environments):
        for elem, value in site:
            elem_len = len([val for val in site if val[0] == elem])
            if elem not in max_num_elem:
                max_num_elem[elem] = 0
            if elem_len > max_num_elem[elem]:
                max_num_elem[elem] = elem_len

    return max_num_elem


def create_site_array(unique_environments, max_num_elems=None, elems=None, normalise=False):
    """ Create padded numpy arrays based on unique environments provided.

    Input:

        | unique_environments: dict(list), dict with element keys full of
                               local substructures.

    Args:

        | max_num_elems : dict(dict(int)), dict with element keys, containing
                          sub-dict with element keys that contain the largest
                          number of each type of element contributing to a site.
                          If None, this will be calculated based on the input
                          environments. The site array is padded to this size.
        | elems         : list(str), custom list of elements to loop over.
        | normalise     : bool, whether to normalise site array to 1.

    Returns:

        | site_array    : dict(np.ndarray), dict with element keys full of
                          normalised site arrays.
        | max_num_elems : dict(dict(int)), dict with element keys containing the
                          highest number of atoms of an element contributing
                          to a particular site, e.g. {'P': {'K': 10}}.

    """
    site_array = defaultdict(list)
    if elems is None:
        elems = [elem for elem in unique_environments]
    if max_num_elems is None:
        max_num_elems = dict()
        for elem in elems:
            max_num_elems[elem] = get_max_coordination_of_elem(unique_environments[elem])
    for elem in elems:
        for site_ind, site in enumerate(unique_environments[elem]):
            site_array[elem].append(dict())
            total_angle = 0
            for _elem in elems:
                if _elem not in max_num_elems[elem]:
                    max_num_elems[elem][_elem] = 0
                site_array[elem][-1][_elem] = np.zeros((max_num_elems[elem][_elem]))
                count = 0
                for species, angle in site:
                    if species == _elem:
                        site_array[elem][-1][_elem][count] = angle
                        count += 1
                        total_angle += angle
            if normalise:
                for _elem in elems:
                    site_array[elem][-1][_elem] /= total_angle

    return site_array, max_num_elems


def set_substruc_dict(doc):
    """ Compute voronoi substructure and collect into dict. Sets the
    'voronoi_substruc' and 'substruc_dict' entries in the doc.

    Input:

        | doc: dict, structure to compute substructure of.

    """
    if 'voronoi_substruc' not in doc:
        from matador.plugins.voronoi_interface.voronoi_interface import get_voronoi_substructure
        doc['voronoi_substruc'] = get_voronoi_substructure(doc)
    voronoi_substruc_dict = dict()
    elems = set(doc['atom_types'])
    for elem in elems:
        voronoi_substruc_dict[elem] = [substruc[1] for substruc in doc['voronoi_substruc'] if substruc[0] == elem]
    doc['substruc_dict'] = voronoi_substruc_dict


def set_site_array(doc, normalise=False):
    """ Set the 'site_array' entry in the chosen document. Creates
    a dict of numpy arrays of normalised solid angles with element keys
    from the import calculated substructure.

    Input:

        | doc: dict, matador doc containing substruc_dict.

    Args:

        | normalise: bool, whether to normalise site_arrays to total angle.

    """
    if 'substruc_dict' not in doc:
        set_substruc_dict(doc)
    doc['site_array'], doc['max_num_elems'] = create_site_array(doc['substruc_dict'], normalise=normalise)


def get_unique_sites(doc, atol=1e-2, rtol=1e-2):
    if 'substruc_dict' not in doc:
        set_substruc_dict(doc)
    elems = set(doc['atom_types'])
    substruc_dict = doc['substruc_dict']
    if 'site_array' not in doc:
        set_site_array(doc)
    site_array = doc['site_array']
    unique_sites = dict()
    degeneracies = dict()
    similar_sites = dict()
    for elem in elems:
        unique_sites[elem] = set()
        similar_sites[elem] = []
        for i in range(len(site_array[elem])):
            for j in range(i+1, len(site_array[elem])):
                same = are_sites_the_same(site_array[elem][i], site_array[elem][j], atol=atol, rtol=rtol)
                if same:
                    added = False
                    for ind, site in enumerate(similar_sites[elem]):
                        if i in site:
                            similar_sites[elem][ind].add(j)
                            added = True
                            break
                        if j in site:
                            similar_sites[elem][ind].add(i)
                            added = True
                            break
                    if not added:
                        similar_sites[elem].append(set([i, j]))
            if not any([i in _set for _set in similar_sites[elem]]):
                similar_sites[elem].append(set([i]))

        # for env in site_array[elem]:
            # print(env)
        from copy import deepcopy
        temp_similar_sites = deepcopy(similar_sites[elem])
        valid = False
        _iter = 0
        while not valid:
            scrubbed = False
            _iter += 1
            if _iter > 100:
                raise RuntimeError
            for index in range(len(site_array[elem])):
                index_sites = []
                for ind, site in enumerate(temp_similar_sites):
                    if index in site:
                        index_sites.append(ind)
                # if index is in multiple sites, combine all sites into the first
                if len(index_sites) > 1:
                    # construct mean of each set of sites, and if they are too far apart, get worried
                    site_mean = []
                    for j in range(len(index_sites)):
                        site_mean.append(defaultdict(list))
                        for _elem in site_array[elem][i]:
                            site_mean[-1][_elem].append(np.mean(np.asarray([site_array[elem][i][_elem] for i in temp_similar_sites[index_sites[j]]]), axis=0))
                            assert np.shape(site_mean[-1][_elem][-1]) == np.shape(site_array[elem][0][_elem])
                    for j in range(len(index_sites)-1, 0, -1):
                        # let means deviate by twice the amount
                        if not are_sites_the_same(site_mean[j], site_mean[0], atol=5*atol, rtol=5*rtol):
                            raise RuntimeError('Numerical instability: site groups with distinct means were clustered.')
                        [temp_similar_sites[index_sites[0]].add(i) for i in list(temp_similar_sites[index_sites[j]])]
                        del temp_similar_sites[index_sites[j]]
                    scrubbed = True
                    break
                elif len(index_sites) == 0:
                    raise RuntimeError
            if not scrubbed:
                valid = True
        similar_sites[elem] = temp_similar_sites

        unique_sites[elem] = [next(iter(site)) for site in similar_sites[elem]]
        degeneracies[elem] = [len(_set) for _set in similar_sites[elem]]

    doc['unique_site_inds'] = unique_sites
    doc['unique_substrucs'] = dict()
    doc['unique_site_array'] = dict()
    doc['unique_site_stddev'] = dict()
    doc['similar_sites'] = similar_sites
    for elem in elems:
        doc['unique_substrucs'][elem] = [substruc_dict[elem][ind] for ind in unique_sites[elem]]
        doc['unique_site_array'][elem] = []
        doc['unique_site_stddev'][elem] = []
        for site in similar_sites[elem]:
            doc['unique_site_array'][elem].append(dict())
            doc['unique_site_stddev'][elem].append(dict())
            for _elem in site_array[elem][0]:
                doc['unique_site_array'][elem][-1][_elem] = np.mean(np.asarray([site_array[elem][i][_elem] for i in site]), axis=0)
                doc['unique_site_stddev'][elem][-1][_elem] = np.std(np.asarray([site_array[elem][i][_elem] for i in site]), axis=0)
    doc['site_degeneracies'] = degeneracies


def compare_docs(docA, docB, elems):
    matching = 0
    print('COMPARING', ' '.join(docA['text_id']), 'vs', ' '.join(docB['text_id']))
    for elem in elems:
        if elem not in docA['unique_substrucs'] or elem not in docB['unique_substrucs']:
            print('Missing elements.')
            return False
        for i, site in enumerate(docA['unique_substrucs'][elem]):
            for other_site in docB['unique_substrucs'][elem]:
                if are_sites_the_same(site, other_site, ['K', 'P']):
                    matching += docA['site_degeneracies'][elem][i]
    print(matching, '/', sum(docA['site_degeneracies'][elem]))
    if matching / sum(docA['site_degeneracies'][elem]) > 0.5:
        return True


def cluster(X, hull, max_of_elems, elems=None, method='KMeans', elem='P', quantile=None, n_clusters=None, cmap='tab10'):
    if method == 'KMeans':
        from sklearn.cluster import KMeans
        inertia = []
        elem = 'P'
        est = KMeans(n_clusters=n_clusters, n_jobs=-2, precompute_distances=True)
        est.fit(X)
        inertia.append(est.inertia_)
        labels = est.labels_
        # means = est.cluster_centers_
        print('Number of clusters: {}'.format(n_clusters))
    elif method == 'MeanShift':
        from sklearn.cluster import MeanShift, estimate_bandwidth
        bandwidth = estimate_bandwidth(X, quantile=quantile if quantile is not None else 0.2)
        est = MeanShift(bandwidth=bandwidth, cluster_all=False)
        est.fit(X)
        labels = est.labels_
        # means = est.cluster_centers_
        n_clusters = len(set(labels))
        print('Number of estimate clusters:', n_clusters)
    """
    means_of_point_class = np.asarray([means[label] for label in labels])
    labels = labels[np.argsort(means_of_point_class[:, 0])]
    X = X[np.argsort(means_of_point_class[:, 0])]
    temp_labels = np.zeros_like(labels)
    sorted_labels = np.zeros_like(labels)
    for i in range(len(labels)-1):
       if labels[i] != labels[i+1]:
           temp_labels[i+1] += 1
    current_group = 0
    for i in range(len(temp_labels)):
       if temp_labels[i] == 1:
           current_group += 1
       sorted_labels[i] = current_group
    """

#    fig, axarr = plt.subplots(2, 2, figsize=(10,10))
#    sns.set_palette(sns.color_palette(cmap, n_colors=n_clusters))
#    colours = [sns.palettes.color_palette(cmap, n_colors=n_clusters)[label] for label in labels]
#
#    sns.stripplot(x=labels, y=plot_X[:, 0], ax=axarr[0, 0], jitter=0.1)
#    axarr[0, 0].set_ylabel('Fraction of solid angle seen as P')
#
#    axarr[0, 1].scatter(plot_X[:, 1], plot_X[:, 3], c=colours, lw=0, alpha=0.3)
#    axarr[0, 1].set_xlabel('Fraction of solid angle seen as P')
#    axarr[0, 1].set_ylabel('Number of contrib P atoms')
#
#    axarr[1, 0].scatter(plot_X[:, 0], plot_X[:, 2], c=colours, lw=0, alpha=0.3)
#    axarr[1, 0].set_xlabel('Fraction of solid angle seen as K')
#    axarr[1, 0].set_ylabel('Number of contrib K atoms')
#
#    sns.stripplot(x=labels, y=plot_X[:, 3], ax=axarr[1, 1], jitter=0.1)
#    axarr[1, 1].set_ylabel('Number of contrib Pf atoms')
#    fig.suptitle('Number of clusters: {}'.format(n_clusters))

    for doc in hull.hull_cursor:
        site_classes = []
        if elem not in doc['unique_substrucs']:
            doc['class'] = -1
            doc['class_err'] = 0
            doc['class_sites'] = [-1]
            continue
        for ind, substruc in enumerate(doc['unique_substrucs'][elem]):
            site_array = np.zeros((1, int(np.sum(max_of_elems[elem]))))
            elem_counters = np.zeros((len(elems)), dtype=int)
            for elem_ind, _elem in enumerate(elems):
                if elem_ind != 0:
                    elem_counters[elem_ind] += max_of_elems[elem][elem_ind-1]
            for elem_ind, _elem in enumerate(elems):
                for atom in substruc:
                    if atom[0] == _elem:
                        site_array[0][elem_counters[elem_ind]] = atom[1]
                        elem_counters[elem_ind] += 1
            site_classes.append(est.predict(site_array))
        site_classes = [val[0] for val in site_classes]
        doc['class'] = np.median(site_classes)
        doc['class_sites'] = [val for val in site_classes]
        doc['class_err'] = (len(set(doc['class_sites']))-1)/n_clusters
        class_vector = np.zeros((n_clusters))
        for site in site_classes:
            class_vector[int(site)] += 1
        doc['class_vector'] = class_vector

    return est


def plot_class_hull(hull, n_clusters, plot_class=None):
    from random import sample
    import matplotlib.pyplot as plt
    fig = plt.figure(figsize=(12, 20))
    ax = fig.add_subplot(211)
    ax2 = fig.add_subplot(212)
    ax2 = hull.plot_2d_hull(show=False, ax=ax2)
    colours = plt.cm.Dark2(np.linspace(0, 1, n_clusters*10))
    class_concs = defaultdict(list)
    for doc in hull.hull_cursor:
        class_concs[int(doc['class'])].append(doc['concentration'])
#        ax.errorbar(doc['concentration'], doc['class'], yerr=doc['class_err'], fmt='o', c=colours[int(10*doc['class'])], alpha=0.01, zorder=100)
    for _class in class_concs:
        ax.errorbar(np.mean(class_concs[_class]), _class, xerr=np.std(class_concs[_class]), fmt='o', c=colours[int(10*_class)], lw=2, zorder=1000)
    bounds = [0.32, 0.57, 0.74]
    for bound in bounds:
        ax.axvline(bound, ls='--', lw=1)
    ax.set_ylim(-1.5, n_clusters)
    ax.set_xlim(-0.1, 1.1)
    for doc in hull.hull_cursor:
        ax2.scatter(doc['concentration'], doc['formation_enthalpy_per_atom'], c='#eeeeee', s=60, zorder=9999, lw=0)
    for doc in sample(hull.hull_cursor, len(hull.hull_cursor)):
            if plot_class is not None:
                if doc['class'] not in plot_class:
                    continue
                else:
                    ax2.scatter(doc['concentration'], doc['formation_enthalpy_per_atom'], c=colours[int(10*(doc['class']))], s=50, alpha=max(1-2*doc['class_err'], 0.1), zorder=99999, lw=0)
            else:
                if doc['hull_distance'] <= 1e-12:
                    ax2.scatter(doc['concentration'], doc['formation_enthalpy_per_atom'], c=colours[int(10*(doc['class']))], s=75, zorder=99999999, lw=1)
                else:
                    ax2.scatter(doc['concentration'], doc['formation_enthalpy_per_atom'], c=colours[int(10*(doc['class']))], s=50, alpha=max(1-2*doc['class_err'], 0.1), zorder=99999, lw=0)


def find_site_index(doc, elem, elem_ind):
    elem_count = 0
    for idx, atom in enumerate(doc['atom_types']):
        if atom == elem:
            if elem_count == elem_ind:
                print('Found substruc at {}'.format(elem_count))
                return elem_count
        elem_count += 1
    return False


def viz_site(doc, targ_substruc, elem, rmax=6):
    if 'positions_abs' not in doc:
        doc['positions_abs'] = frac2cart(doc['lattice_cart'], doc['positions_frac'])
    for ind, substruc in enumerate(doc['substruc_dict'][elem]):
        if substruc == targ_substruc:
            elem_ind = ind
            elem_count = find_site_index(doc, elem, elem_ind)
            if elem_count is not False:
                break
    targ_site = doc['positions_frac'][elem_count]
    targ_pos = doc['positions_abs'][elem_count]
    print(elem_count)
    from itertools import product
    from collections import defaultdict
    doc['lattice_cart'] = np.asarray(doc['lattice_cart'])
    neighbour_doc = defaultdict(list)
    neighbour_doc['positions_abs'].append(targ_pos)
    neighbour_doc['positions_frac'].append(targ_site)
    neighbour_doc['atom_types'].append('U')
    neighbour_doc['lattice_cart'] = doc['lattice_cart']
    for j, pos in enumerate(doc['positions_abs']):
        for prod in product(range(-1, 2), repeat=3):
            trans = np.zeros((3))
            for ind, multi in enumerate(prod):
                trans += doc['lattice_cart'][ind] * multi
            dist2 = 0
            for i in range(3):
                dist2 += (targ_pos[i]-(pos+trans)[i])**2
            if dist2 < rmax**2:
                neighbour_doc['positions_abs'].append(pos + trans)
                neighbour_doc['positions_frac'].append((np.asarray(doc['positions_frac'][j]) + prod).tolist())
                neighbour_doc['atom_types'].append(doc['atom_types'][j])
    print(targ_site)
    return elem_count, neighbour_doc



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the abstract Calculator class for use
with compute.

"""
import abc
from matador.utils.errors import CalculationError


class Calculator:
    """ Base class for Calculator objects, with some useful methods and
    abstract methods.

    """
    def __init__(self, calculation_parameters):
        self._initial_calculation_parameters = {}
        self._task = calculation_parameters['task']

    @staticmethod
    @abc.abstractmethod
    def verify_calculation_parameters(calculation_parameters, structure):
        """ Verify if the parameters passed make sense for this calculator.
        Mutates the input dict if required.

        Parameters:
            calculation_parameters (dict): contains global parameters, i.e.
                non-structure-specific parameters.
            structure (dict): contains the structure.

        Raises:
            InputError: if verification fails, with a (hopefully) helpful message.

        """
        return

    @staticmethod
    def verify_simulation_cell(structure):
        """ Run some checks on the structure.

        Parameters:
            structure (dict): dictionary containing the structure.

        Raises:
            CalculationError: if cell is pathological.

        """
        errors = []
        if all([angle < 30 for angle in structure['lattice_abc'][1]]):
            msg = 'Cell is pathological (at least one angle < 30).'
            errors.append(msg)

        # TODO: check overlapping atoms

        if errors:
            raise CalculationError('. '.join(errors))

    def do_memcheck(self):
        """ If possible, do a dryrun of the code to estimate memory usage.

        Returns:
            int: estimated memory usage in MB. 0 if method not implemented.

        """
        # TODO: reimplement
        return 0



#!/usr/bin/env python
import unittest
from matador.fingerprints.pxrd import PXRD
from matador.scrapers.castep_scrapers import res2dict
import numpy as np
from os.path import realpath

REAL_PATH = "/".join(realpath(__file__).split("/")[:-1]) + "/"
DEBUG = True
PEAK_POS_TOL = 0.02


def _match_peaks(res_file, gsas_reflections, **kwargs):
    """ Check if peaks match with HKL between matador and GSAS. """
    doc, s = res2dict(res_file, as_model=True)

    pxrd = PXRD(doc, **kwargs)
    gsas_peaks = np.loadtxt(gsas_reflections)
    hkl_array = {}
    for peak in gsas_peaks:
        hkl_array[(int(peak[0]), int(peak[1]), int(peak[2]))] = peak[3]

    for ind, hkl in enumerate(pxrd.hkls):
        hkl_tuple = (int(hkl[0]), int(hkl[1]), int(hkl[2]))
        if hkl_tuple in hkl_array:
            assert (
                np.abs(hkl_array[hkl_tuple] - pxrd.peak_positions[ind]) < PEAK_POS_TOL
            )

    return pxrd


class PXRDCalculatorTest(unittest.TestCase):
    """ Test PXRD calculator. """

    def test_simple_pxrd(self):
        """ Test Li PXRD vs GSAS. """
        _ = _match_peaks(
            REAL_PATH + "data/structures/Li.res",
            REAL_PATH + "data/pxrd_files/Li_reflections_Cu.txt",
        )
        pxrd_args = {"wavelength": 0.559363}
        _ = _match_peaks(
            REAL_PATH + "data/structures/Li.res",
            REAL_PATH + "data/pxrd_files/Li_reflections_Ag.txt",
            **pxrd_args
        )

    def test_CuP2_vs_GSAS(self):
        """ Test CuP2 peak positions vs GSAS. """
        pxrd = _match_peaks(
            REAL_PATH + "data/pxrd_files/CuP2.res",
            REAL_PATH + "data/pxrd_files/CuP2_GSASII_reflections.txt",
        )
        self.assertAlmostEqual(
            pxrd.two_thetas[np.argmax(pxrd.pattern)], 30.969, places=2
        )



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements plotting routines specifically
for the PDF and PDFOverlap objects defined in the
matador.fingerprints.pdf module.

"""

import numpy as np

from matador.fingerprints.pdf import PDF
from matador.crystal import Crystal
from matador.plotting.plotting import plotting_function
from matador.utils.cell_utils import get_space_group_label_latex
from matador.utils.chem_utils import get_formula_from_stoich

__all__ = ['plot_pdf', 'plot_projected_pdf', 'plot_diff_overlap', 'plot_projected_diff_overlap']


@plotting_function
def plot_pdf(pdfs,
             labels=None, r_min=None, r_max=None,
             offset=1.2, text_offset=(0.0, 0.0),
             legend=False, annotate=True, figsize=None,
             filename=None,
             **kwargs):
    """ Plot PDFs.

    Parameters:
        pdfs (list of matador.fingerprints.pdf.PDF or matador.crystal.Crystal or dict):
            the PDFs to plot, as a list of PDF or Crystal objects, or a matador document.

    Keyword arguments:
        labels (list of str): labels to add to the PDF plot.
        offset (float): amount by which to separate the PDFs in the plot. A value of 1
            will separate by the maximum intensity across the PDFs. Default is 1.5.
        text_offset (tuple of float): two float values to move annotations around relative
            to the base of their corresponding PDF, in units of (Angstrom, max_gr).
        r_max (float): the radius to plot out to. Default is the minmax(radius across
            all PDFs).
        annotate (bool): whether or not to apply the PDF labels as an annotation.
        legend (bool): whether or not to apply the PDF labels as a legend.
        figsize (tuple of float): matplotlib figure size. Default scales with number of PDFs.

    Returns:
        matplotlib.pyplot.Axes: axis object which can be modified further.

    """

    import matplotlib.pyplot as plt

    if not isinstance(pdfs, list):
        pdfs = [pdfs]
    if labels is not None and not isinstance(labels, list):
        labels = [labels]

    if figsize is None:
        _user_default_figsize = plt.rcParams.get('figure.figsize', (8, 6))
        height = len(pdfs) * max(0.5, _user_default_figsize[1] / 1.5 / len(pdfs))
        figsize = (_user_default_figsize[0], height)

    fig = plt.figure(figsize=figsize)
    ax1 = fig.add_subplot(111)

    if labels is not None and len(labels) != len(pdfs):
        raise RuntimeError("Wrong number of labels {} for PDFs.".format(labels))

    if isinstance(pdfs[0], Crystal):
        gr_max = max(np.max(pdf.pdf.gr) for pdf in pdfs)
        _r_max = min(np.max(pdf.pdf.r_space) for pdf in pdfs)
    elif isinstance(pdfs[0], dict):
        gr_max = max(np.max(pdf['pdf'].gr) for pdf in pdfs)
        _r_max = min(np.max(pdf['pdf'].r_space) for pdf in pdfs)
    else:
        gr_max = max(np.max(pdf) for pdf in pdfs)
        _r_max = min(np.max(pdf.r_space) for pdf in pdfs)
    abs_offset = offset * gr_max

    if r_max is None:
        r_max = _r_max
    if r_min is None:
        r_min = 0.0

    ax1.set_ylabel('Pair distribution function, $g(r)$')
    ax1.get_yaxis().set_ticks([])
    ax1.set_xlim(r_min, r_max+0.5)

    for ind, pdf in enumerate(pdfs):

        if isinstance(pdf, Crystal):
            pdf = pdf.pdf
        elif isinstance(pdf, dict) and 'pdf' in pdf:
            pdf = pdf['pdf']

        if labels:
            label = labels[ind]
        else:
            label = get_space_group_label_latex(pdf.spg) + '-' + get_formula_from_stoich(pdf.stoichiometry, tex=True)

        ax1.plot(pdf.r_space, pdf.gr + abs_offset * ind, label=label)
        if text_offset is not None:
            text_x = text_offset[0]
        if text_offset is not None:
            text_y = abs_offset*ind + text_offset[1]*gr_max
        if label is not None and annotate:
            ax1.text(text_x, text_y, label)

    ax1.set_ylim(-gr_max * 0.2, offset * gr_max * len(pdfs))

    ax1.set_xlabel('$r$ ($\\AA$)')

    if legend:
        legend = ax1.legend()

    if any([kwargs.get('pdf'), kwargs.get('svg'), kwargs.get('png')]):
        bbox_extra_artists = None
        if filename is None:
            filename = '-'.join([get_formula_from_stoich(pdf.stoichiometry) for pdf in pdfs]) + '_pdf'

        if kwargs.get('pdf'):
            plt.savefig('{}.pdf'.format(filename),
                        bbox_inches='tight', transparent=True, bbox_extra_artists=bbox_extra_artists)
        if kwargs.get('svg'):
            plt.savefig('{}.svg'.format(filename),
                        bbox_inches='tight', transparent=True, bbox_extra_artists=bbox_extra_artists)
        if kwargs.get('png'):
            plt.savefig('{}.png'.format(filename),
                        bbox_inches='tight', transparent=True, bbox_extra_artists=bbox_extra_artists)

    return ax1


@plotting_function
def plot_projected_pdf(pdf, keys=None, other_pdfs=None, vlines=None):
    """ Plot projected PDFs.

    Parameters:
        pdf (matador.fingerprints.pdf.PDF): the main PDF to plot.

    Keyword arguments:
        keys (list): plot only a subset of projections, e.g. [('K', )].
        other_pdfs (list of PDF): other PDFs to plot.
        vlines (list of float): plot vertical lines at these points.

    """
    import matplotlib.pyplot as plt
    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    if keys is None:
        keys = [key for key in pdf.elem_gr]
    for key in keys:
        if key not in pdf.elem_gr:
            key = (key[1], key[0])
        ax1.plot(pdf.r_space, pdf.elem_gr[key], label='-'.join(key) + ' {}'.format(pdf.label))
    if other_pdfs is not None:
        if isinstance(other_pdfs, PDF):
            other_pdfs = [other_pdfs]
        for _pdf in other_pdfs:
            if isinstance(_pdf, PDF):
                for key in keys:
                    ax1.plot(_pdf.r_space, _pdf.elem_gr[key], ls='--',
                             label='-'.join(key) + ' {}'.format(_pdf.label))
            elif isinstance(pdf, tuple):
                ax1.plot(_pdf[0], _pdf[1], alpha=1, ls='--')
            else:
                raise RuntimeError

    if vlines is not None:
        for line in vlines:
            ax1.axvline(line, ls='--', alpha=0.8, c='grey')
    ax1.legend(loc=1)
    ax1.set_ylabel('$g(r)$')
    ax1.set_xlabel('$r$ (Angstrom)')


@plotting_function
def plot_diff_overlap(pdf_overlap):
    """ Simple plot for comparing two PDFs.

    Parameters:
        pdf_overlap (matador.fingerprints.pdf.PDFOverlap): the
        overlap object to plot.

    """
    import matplotlib.pyplot as plt
    import matplotlib.gridspec as gridspec
    import numpy as np

    plt.figure()
    gs = gridspec.GridSpec(2, 1, height_ratios=[2, 1])
    gs.update(hspace=0)

    ax1 = plt.subplot(gs[0])
    ax2 = plt.subplot(gs[1], sharex=ax1)

    ax2.set_xlabel('$r$ (\\AA)')
    ax1.set_ylabel('$g(r)$')
    ax2.set_ylabel('$g_a(r) - g_b(r)$')
    ax2.axhline(0, ls='--', c='k', lw=0.5)
    ax1.set_xlim(0, np.max(pdf_overlap.fine_space))

    ax1.plot(pdf_overlap.fine_space, pdf_overlap.fine_gr_a, label=pdf_overlap.pdf_a.label)
    ax1.plot(pdf_overlap.fine_space, pdf_overlap.fine_gr_b, label=pdf_overlap.pdf_b.label)

    plt.setp(ax1.get_xticklabels(), visible=False)
    ax2.set_ylim(-0.5 * ax1.get_ylim()[1], 0.5 * ax1.get_ylim()[1])

    ax1.legend(loc=0)
    ax2.plot(pdf_overlap.fine_space, pdf_overlap.overlap_fn, ls='-')
    ax2.set_ylim(ax1.get_ylim()[1], ax1.get_ylim()[1])


@plotting_function
def plot_projected_diff_overlap(pdf_overlap):
    """ Simple plot for comparing two PDFs.

    Parameters:
        pdf_overlap (matador.fingerprints.pdf.PDFOverlap): the
            overlap object to plot.

    """
    import matplotlib.pyplot as plt
    import numpy as np
    import matplotlib.gridspec as gridspec

    plt.figure()
    gs = gridspec.GridSpec(2, 1, height_ratios=[2, 1])
    gs.update(hspace=0)

    ax1 = plt.subplot(gs[0])
    ax2 = plt.subplot(gs[1], sharex=ax1)
    ax2.set_xlabel('$r$ (\\AA)')
    ax1.set_ylabel('$g(r)$')
    ax2.set_ylabel('$g_a(r) - g_b(r)$')
    ax2.axhline(0, ls='--', c='k', lw=0.5)
    ax1.set_xlim(0, np.max(pdf_overlap.fine_space))
    for _, key in enumerate(pdf_overlap.fine_elem_gr_a):
        ax1.plot(pdf_overlap.fine_space, pdf_overlap.fine_elem_gr_a[key],
                 label='-'.join(key) + ' {}'.format(pdf_overlap.pdf_a.label))
        ax1.plot(pdf_overlap.fine_space, pdf_overlap.fine_elem_gr_b[key],
                 label='-'.join(key) + ' {}'.format(pdf_overlap.pdf_b.label),
                 ls='--')
        ax2.plot(pdf_overlap.fine_space, pdf_overlap.fine_elem_gr_a[key] - pdf_overlap.fine_elem_gr_b[key],
                 label='-'.join(key) + ' diff')
    plt.setp(ax1.get_xticklabels(), visible=False)
    ax2.set_ylim(ax1.get_ylim()[1], ax1.get_ylim()[1])
    ax1.legend(loc=0)
    ax2.legend(loc=2)



""" Useful base classes for tests. """

import unittest
import os
import subprocess as sp
import shutil

REAL_PATH = "/".join(os.path.realpath(__file__).split("/")[:-1]) + "/"
TMP_DIR = "tmp_test"
ROOT_DIR = os.getcwd()


def detect_program(executable):
    try:
        with open("/dev/null", "w") as devnull:
            out, errs = sp.Popen(
                [executable, "--version"], stdout=devnull, stderr=devnull
            ).communicate()
        if errs:
            raise RuntimeError
        return True
    except Exception:
        return False


class MatadorUnitTest(unittest.TestCase):
    def tearDown(self):
        os.chdir(REAL_PATH)
        if os.path.isdir(TMP_DIR):
            shutil.rmtree(TMP_DIR)

        os.chdir(ROOT_DIR)

    def setUp(self):
        os.chdir(REAL_PATH)
        os.makedirs(TMP_DIR, exist_ok=True)
        os.chdir(TMP_DIR)



# coding: utf-8
# Distributed under the terms of the MIT License.

from pathlib import Path

from .atomic_scattering_loaders import load_scattering_file

__all__ = ["RASPA_ATOMIC_SCATTERING_COEFFS"]


def load_atomic_scattering_factors_raspa():
    """ Load atomic scattering factors used by RASPA2, with unknown
    reference.

    https://github.com/numat/RASPA2/blob/master/src/scattering_factors.c

    """
    filename = Path(__file__).parent.joinpath("atomic_scattering_factors.dat")
    return load_scattering_file(
        filename, a_inds=(1, 3, 5, 7), b_inds=(2, 4, 6, 8), c_ind=9
    )


RASPA_ATOMIC_SCATTERING_COEFFS = load_atomic_scattering_factors_raspa()



# coding: utf-8
# Distributed under the terms of the MIT license.

""" This file implements classes to store and manipulate magnetic
resonance data.

"""

from matador.orm.orm import DataContainer


class MagneticResonance(DataContainer):
    pass



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements the "matador" command line. """


from matador.query import DBQuery
from matador.hull import QueryConvexHull
from matador.utils.print_utils import print_failure, print_warning, print_notify
from matador.db import make_connection_to_collection
from matador.config import load_custom_settings


class MatadorCommandLine:
    """ Class that implements the command-line interface to a MongoDB
    structure repository.

    """
    def __init__(self, *args, **kwargs):
        """ Initialise the query with command line arguments and return
        results.

        """
        # read args
        self.kwargs = kwargs
        self.args = vars(args[0])
        self.args['no_quickstart'] = self.kwargs.get('no_quickstart')
        self.argstr = kwargs.get('argstr')

        file_exts = ['cell', 'res', 'pdb', 'markdown', 'latex', 'param', 'xsf']
        self.export = any([self.args.get(ext) for ext in file_exts])

        self.subcommand = self.args.pop("subcmd")

        if self.subcommand != 'import':
            self.settings = load_custom_settings(config_fname=self.args.get('config'),
                                                 debug=self.args.get('debug'),
                                                 no_quickstart=self.args.get('no_quickstart'))
            result = make_connection_to_collection(self.args.get('db'),
                                                   check_collection=(self.subcommand != "stats"),
                                                   mongo_settings=self.settings)
            self.client, self.db, self.collections = result

        if self.subcommand == 'stats':
            self.stats()

        try:
            if self.subcommand == 'import':
                from matador.db import Spatula
                self.importer = Spatula(self.args)

            if self.subcommand == 'query':
                self.query = DBQuery(self.client, self.collections, **self.args)
                self.cursor = self.query.cursor

            if self.subcommand == 'swaps':
                from matador.swaps import AtomicSwapper
                self.query = DBQuery(self.client, self.collections, **self.args)
                if self.args.get('hull_cutoff') is not None:
                    self.hull = QueryConvexHull(query=self.query, **self.args)
                    self.swapper = AtomicSwapper(self.hull.hull_cursor, **self.args)
                else:
                    self.swapper = AtomicSwapper(self.query.cursor, **self.args)
                self.cursor = self.swapper.cursor

            if self.subcommand == 'refine':
                from matador.db import Refiner
                self.query = DBQuery(self.client, self.collections, **self.args)
                if self.args.get('hull_cutoff') is not None:
                    self.hull = QueryConvexHull(self.query, **self.args)
                    self.refiner = Refiner(self.hull.cursor, self.query.repo, **self.args)
                else:
                    self.refiner = Refiner(self.query.cursor, self.query.repo, **self.args)

                self.cursor = self.refiner.cursor

            if self.subcommand == 'hull' or self.subcommand == 'voltage':
                self.hull = QueryConvexHull(**self.args,
                                            voltage=self.subcommand == 'voltage',
                                            client=self.client,
                                            collections=self.collections)
                self.cursor = self.hull.hull_cursor

            if self.subcommand == 'changes':
                from matador.db import DatabaseChanges
                if len(self.collections) != 1:
                    raise SystemExit('Cannot view changes of more than one collection at once.')
                if self.args.get('undo'):
                    action = 'undo'
                else:
                    action = 'view'
                changeset = self.args.get('changeset')
                if changeset is None:
                    changeset = 0
                DatabaseChanges([key for key in self.collections][0],
                                changeset_ind=changeset,
                                action=action,
                                mongo_settings=self.settings,
                                override=kwargs.get('no_quickstart'))

            if self.subcommand == 'hulldiff':
                from matador.hull.hull_diff import diff_hulls
                if self.args.get('compare') is None:
                    raise SystemExit('Please specify which hulls to query with --compare.')
                diff_hulls(self.client, self.collections, **self.args)

            if self.export and self.cursor:
                from matador.export import query2files
                if self.args.get('write_n') is not None:
                    self.cursor = [doc for doc in self.cursor if len(doc['stoichiometry']) == self.args.get('write_n')]
                if not self.cursor:
                    print_failure('No structures left to export.')
                query2files(self.cursor, **self.args, argstr=self.argstr, subcmd=self.subcommand, hash_dupe=True)

            if self.args.get('view'):
                from matador.utils.viz_utils import viz
                if self.args.get('top') is None:
                    self.top = len(self.cursor)
                else:
                    self.top = self.args.get('top')
                if len(self.cursor[:self.top]) > 10:
                    from time import sleep
                    print_warning('WARNING: opening {} files with ase-gui...'.format(len(self.cursor)))
                    print_warning('Please kill script within 3 seconds if undesired...')
                    sleep(3)
                if len(self.cursor[:self.top]) > 20:
                    print_failure(
                        'You will literally be opening that many windows, ' +
                        'I\'ll give you another 5 seconds to reconsider...')

                    sleep(5)
                    print_notify('It\'s your funeral...')
                    sleep(1)
                for doc in self.cursor[:self.top]:
                    viz(doc)

            if self.subcommand != 'import':
                self.client.close()

        except (RuntimeError, SystemExit, KeyboardInterrupt) as oops:
            if isinstance(oops, RuntimeError):
                print_failure(oops)
            elif isinstance(oops, SystemExit):
                print_warning(oops)
            try:
                self.client.close()
            except AttributeError:
                pass
            raise oops

    def print_report(self):
        """ Print spatula report on current database. """
        try:
            report = self.report.find_one()
            print('Database last modified on', report['last_modified'], 'with matador', report['version'] + '.')
        except Exception:
            print_warning('Failed to print database report: spatula is probably running!')

    def stats(self):
        """ Print some useful stats about the database. """
        if self.args.get('list'):
            print_notify(str(len(self.db.list_collection_names())) + ' collections found in database:\n')
            collstats_list = []
            for name in self.db.list_collection_names():
                collstats_list.append(self.db.command('collstats', name))
                collstats_list[-1]['name'] = name
            collstats_list = sorted(collstats_list, key=lambda k: k['count'], reverse=True)
            print("\t{:^20}\t{:^20}".format('Name', 'Number of structures'))
            for collection in collstats_list:
                if not collection['name'].startswith('__'):
                    print("\t{:<20}\t{:>20d}".format(collection['name'], collection['count']))
            print('\n')
        elif self.args.get('delete'):
            target = self.args.get('db')
            if isinstance(target, list) and len(target) == 1:
                target = target[0]
            else:
                raise SystemExit('I will only delete one collection at a time...')
            if target is None:
                raise SystemExit('Please specify a collection to delete.')
            if target not in self.db.list_collection_names():
                raise SystemExit('No collection named {} was found'.format(target))

            from getpass import getuser
            user = getuser()
            if user not in target:
                raise SystemExit('I cannot delete a collection that\'s name does not start with '
                                 'your username, {}'.format(user))
            stats = self.db.command('collstats', target)

            if self.args.get('no_quickstart'):
                answer = 'y'
            else:
                answer = input('Are you sure you want to delete collection {} containing {} '
                               'structures? [y/n]\n'.format(target, stats['count']))
            if answer.lower() == 'y':
                if target == 'repo':
                    raise SystemExit('I\'m sorry Dave, I\'m afraid I can\'t do that...')
                print('Deleting collection {}...'.format(target))
                self.db[target].drop()
                print('and its changelog...')
                self.db['__changelog_{}'.format(target)].drop()
            else:
                raise SystemExit('Nevermind then!')

        else:
            comp_list = dict()
            stats_dict = dict()
            stats_dict['count'] = 0
            stats_dict['avgObjSize'] = 0
            stats_dict['storageSize'] = 0
            stats_dict['totalIndexSize'] = 0
            for collection in self.collections:
                db_stats_dict = self.db.command('collstats', collection)
                stats_dict['count'] += db_stats_dict['count']
                stats_dict['avgObjSize'] += db_stats_dict['avgObjSize']
                stats_dict['storageSize'] += db_stats_dict['storageSize']
                stats_dict['totalIndexSize'] += db_stats_dict['totalIndexSize']
            print(("The collection(s) queried in {} contain {} structures at {:.1f} kB each "
                   "totalling {:.1f} MB with a further {:.1f} MB of indexes.")
                  .format(self.db.name,
                          stats_dict['count'],
                          stats_dict['avgObjSize'] / (1024),
                          stats_dict['storageSize'] / (1024**2),
                          stats_dict['totalIndexSize'] / (1024**2)))
            for collname in self.collections:
                cursor = self.collections[collname].find()
                for doc in cursor:
                    temp = ''
                    for ind, elem in enumerate(sorted(doc['stoichiometry'])):
                        temp += str(elem[0])
                        if ind != len(doc['stoichiometry']) - 1:
                            temp += '+'
                    if temp not in comp_list:
                        comp_list[temp] = 0
                    comp_list[temp] += 1
            keys = list(comp_list.keys())
            vals = list(comp_list.values())
            comp_list = list(zip(keys, vals))
            comp_list.sort(key=lambda t: t[1], reverse=True)
            small_count = 0
            first_ind = 1000
            cutoff = 100
            for ind, comp in enumerate(comp_list):
                if comp[1] < cutoff:
                    if ind < first_ind:
                        first_ind = ind
                    small_count += comp[1]
            comp_list = comp_list[:first_ind]
            comp_list.append(['others < ' + str(cutoff), small_count])
            comp_list.sort(key=lambda t: t[1], reverse=True)
            try:
                from ascii_graph import Pyasciigraph
                from ascii_graph.colors import Gre, Blu, Red
                from ascii_graph.colordata import hcolor
            except ImportError:
                print("ascii_graph dependency not found, not creating histogram.")
            else:
                graph = Pyasciigraph(line_length=80, multivalue=False)
                thresholds = {int(stats_dict['count'] / 40): Gre,
                              int(stats_dict['count'] / 10): Blu,
                              int(stats_dict['count'] / 4): Red}
                data = hcolor(comp_list, thresholds)
                for line in graph.graph(label=None, data=data):
                    print(line)
                print('\n')

                for comp in comp_list:
                    print(comp)


def main(no_quickstart=False):
    """ Parse all user args and construct a MatadorCommandLine object.

    Keyword arguments:
        no_quickstart: no_quickstart all stdin with sensible defaults.

    """
    import argparse
    from sys import argv
    from matador import __version__, script_epilog

    parser = argparse.ArgumentParser(
        prog='matador',
        description='MATerial and Atomistic Database Of Refined structures.',
        epilog=script_epilog
    )
    parser.add_argument('--version', action='version', version='matador version ' + __version__ + '.')

    # define subparsers for self.subcommands
    subparsers = parser.add_subparsers(title='self.subcommands', description='valid sub-commands', dest='subcmd')

    # define parent parser for global arguments
    global_flags = argparse.ArgumentParser(add_help=False)

    # common arguments to all self.subcommands
    global_flags.add_argument('--db', nargs='+', help='choose which collection to query')
    global_flags.add_argument('--debug', action='store_true', help='enable debug printing throughout code.')
    global_flags.add_argument('-conf', '--config', type=str,
                              help='specify custom location of matador config file.'
                                   '(DEFAULT: $MATADOR_ROOT/config/matador_conf.json)')
    global_flags.add_argument('--devel', action='store_true', help='test devel code.')
    global_flags.add_argument('--profile', action='store_true', help='run code profiler.')
    global_flags.add_argument('-q', '--quiet', action='store_true', help='redirect most output to /dev/null.')

    # define all other flags by group
    structure_flags = argparse.ArgumentParser(add_help=False)
    structure_flags.add_argument('-c', '--composition', type=str, nargs='+',
                                 help='find all structures containing exclusively the given '
                                      'elements, e.g. LiSi. Macros defined for groups [I]-[VII] '
                                      '[Tran] [Lan] and [Act], used with square brackets.')
    structure_flags.add_argument('-int', '--intersection', action='store_true',
                                 help='query the intersection of compositions instead of the union '
                                      'e.g. -c LiSnS -int queries Li, Sn, S, LiSn, LiS & LiSnS.')
    structure_flags.add_argument('-n', '--num_species', type=int,
                                 help='find all structures containing a certain number of species.')
    structure_flags.add_argument('-f', '--formula', type=str, nargs='+',
                                 help='query a particular chemical formula, e.g. GeTeSi3')
    structure_flags.add_argument('-i', '--id', type=str, nargs='+',
                                 help='specify a particular structure by its text_id')
    structure_flags.add_argument('-ac', '--calc_match', action='store_true',
                                 help='display calculations of the same accuracy as specified id')
    structure_flags.add_argument('-kpttol', '--kpoint_tolerance', type=float,
                                 help='kpoint tolerance for calculation matches (DEFAULT: +/- 0.01 1/Å)')
    structure_flags.add_argument('-presstol', '--pressure_tolerance', type=float,
                                 help='pressure tolerance for calculation matches (DEFAULT: +/- 0.5 GPa)')
    structure_flags.add_argument('-z', '--num_fu', type=int,
                                 help='query a calculations with more than n formula units')
    structure_flags.add_argument('-sg', '--space_group', help='query a particular space group')
    structure_flags.add_argument('-u', '--uniq', type=float, nargs='?', const=0.1,
                                 help='float, return only unique structures (filtered by PDF '
                                      'overlap), to this tolerance (DEFAULT: 0.1)')
    structure_flags.add_argument('-p', '--pressure', type=float,
                                 help='specify an isotropic external pressure to search for, e.g. 10 (GPa)')
    structure_flags.add_argument('-pf', '--partial-formula', action='store_true',
                                 help='stoichiometry/composition queries will include other unspecified species, e.g. '
                                      '-pf search for Li will query any structure containing Li, not just pure Li.')
    structure_flags.add_argument('--tags', nargs='+', type=str, help=('search for manual tags'))
    structure_flags.add_argument('--doi', type=str, help=('search for DOI in format xxxx/xxxx'))
    structure_flags.add_argument('-icsd', '--icsd', type=int, const=0, nargs='?', help=('search for an ICSD CollCode'))
    structure_flags.add_argument('-ss', '--src_str', type=str,
                                 help=('search for a string inside the structure sources'))
    structure_flags.add_argument('-root', '--root_src', type=str,
                                 help=('search for a root_source string of the structure'))
    structure_flags.add_argument('-encap', '--encapsulated', action='store_true',
                                 help='query only structures encapsulated in a carbon nanotube.')
    structure_flags.add_argument('-cntr', '--cnt_radius', type=float,
                                 help='specify the radius of the encapsulating nanotube to within 0.01 Å')
    structure_flags.add_argument('-cntv', '--cnt_vector', type=int, nargs='+',
                                 help='specify the chiral vector of the encapsulating nanotube')
    structure_flags.add_argument('-ecut', '--cutoff', type=float, nargs='+',
                                 help='specify the min. and optionally max. planewave cutoff.')
    structure_flags.add_argument('-geom', '--geom_force_tol', type=float, nargs='+',
                                 help='force tolerance in eV/Å to query for calc matches.')
    structure_flags.add_argument('-grid', '--grid_scale', type=float, nargs='+',
                                 help='grid scale to query for calc matches.')
    structure_flags.add_argument('-finegrid', '--fine_grid_scale', type=float, nargs='+',
                                 help='fine grid scale to query for calc matches.')
    structure_flags.add_argument('--sedc', type=str, help='specify the dispersion correction scheme, e.g. TS or null.')
    structure_flags.add_argument('-xc', '--xc_functional', type=str,
                                 help='specify an xc-functional to query (case-insensitive).')
    structure_flags.add_argument('-kpts', '--mp_spacing', type=float,
                                 help='specify an MP grid spacing in 2π/Å units, e.g. 0.05, will return all values '
                                      'structures with value within --kpt_tol')
    structure_flags.add_argument('--spin', type=str,
                                 help='specifiy whether to query non-spin-polarized (0) calcs or spin polarized calcs '
                                      '(!=1), or lump them both together with `any`')
    structure_flags.add_argument('--loose', action='store_true',
                                 help='loosely matches with calc_match, i.e. only matches pspot and xc_functional')
    structure_flags.add_argument('--ignore_warnings', action='store_true', help='includes possibly bad structures')
    structure_flags.add_argument('--field', type=str, action='append',
                                 help='name of arbitrary field to query')
    structure_flags.add_argument('--filter', nargs='+', action='append',
                                 help='specify either float [min, max] or a string/float value.')

    material_flags = argparse.ArgumentParser(add_help=False)
    material_flags.add_argument('-hc', '--hull_cutoff', type=float,
                                help='return only structures within a certain distance from hull in eV/atom')
    material_flags.add_argument('-lc', '--label_cutoff', nargs='+', type=float,
                                help='label only structures within a certain distance from hull in eV/atom')
    material_flags.add_argument('--biggest', action='store_true',
                                help='use the largest subset of structures to create a hull')
    material_flags.add_argument('--volume', action='store_true',
                                help='plot a volume curve from convex hull (currently limited to binaries)')
    material_flags.add_argument('--chempots', type=float, nargs='+',
                                help='manually specify chem pots as enthalpy per atom for a rough hull.')

    plot_flags = argparse.ArgumentParser(add_help=False)
    plot_flags.add_argument('--pdf', action='store_true', help='save pdf rather than showing plot in X')
    plot_flags.add_argument('--png', action='store_true', help='save png rather than showing plot in X')
    plot_flags.add_argument('--csv', action='store_true', help='save plotting data to separate csv files')
    plot_flags.add_argument('--labels', action='store_true', help='label hull plots')
    plot_flags.add_argument('--svg', action='store_true', help='save svg rather than showing plot in X')
    plot_flags.add_argument('--subplot', action='store_true', help='plot combined hull and voltage graph')
    plot_flags.add_argument('--no_plot', action='store_true', help='suppress plotting')
    plot_flags.add_argument('--capmap', action='store_true', help='plot heat map of gravimetric capacity')
    plot_flags.add_argument('--sampmap', action='store_true', help='plot heat map of concentration sampling')
    plot_flags.add_argument('--efmap', action='store_true', help='plot heat map of formation energy')
    plot_flags.add_argument('--pathways', action='store_true',
                            help='plot line from stable B_x C_y to pure A in ABC ternary.')
    plot_flags.add_argument('--expt', type=str, help='enter experimental voltage curve .csv file for plotting.')
    plot_flags.add_argument('--expt_label', type=str, help='label for experimental data on voltage curve.')

    import_flags = argparse.ArgumentParser(add_help=False)
    import_flags.add_argument('-d', '--dryrun', action='store_true',
                              help='run the importer without connecting to the database')
    import_flags.add_argument('-v', '--verbosity', type=int, help='enable verbose output', default=0)
    import_flags.add_argument('-f', '--force', action='store_true', help='override main database protection')
    import_flags.add_argument('-t', '--tags', nargs='+', type=str, help='set user tags, e.g. nanotube, project name')
    import_flags.add_argument('--recent_only', action='store_true', help='sort files by creation date (st_ctime) and '
                              'stop importing after a duplicate is found in the database.')
    import_flags.add_argument('-s', '--scan', action='store_true',
                              help='only scan the database for new structures, do not import new structures')
    import_flags.add_argument('-p', '--prototype', action='store_true',
                              help='create a database of prototype structures that contain no DFT calculations')

    changes_flags = argparse.ArgumentParser(add_help=False)
    changes_flags.add_argument('-c', '--changeset', type=int, help='changeset number to query')
    changes_flags.add_argument('-r', '--revert', type=int, help='revert database to specified changeset')
    changes_flags.add_argument('-u', '--undo', action='store_true', help='undo changeset')

    collection_flags = argparse.ArgumentParser(add_help=False)
    collection_flags.add_argument('--to', type=str, help='the text_id of a structure with the desired parameters')
    collection_flags.add_argument('--with', type=str,
                                  help=('the seedname (must be within pwd) of cell and param ' +
                                        'files to use for swaps'))
    collection_flags.add_argument('--prefix', type=str,
                                  help='add a prefix to all file names to write out (auto-appended with an underscore')

    query_flags = argparse.ArgumentParser(add_help=False)

    query_flags.add_argument('-s', '--summary', action='store_true',
                             help='show only the ground state for each stoichiometry.')
    query_flags.add_argument('-t', '--top', type=int, help='number of structures to show/write (DEFAULT: 10)')
    query_flags.add_argument('-dE', '--delta_E', type=float,
                             help='maximum distance from ground state structure to show/write in eV/atom')
    query_flags.add_argument('-d', '--details', action='store_true',
                             help='show as much detail about calculation as possible')
    query_flags.add_argument('-pa', '--per_atom', action='store_true', help='show quantities per atom not per fu.')
    query_flags.add_argument('-ef', '--eform', action='store_true', help='print formation energy not hull distance.')
    query_flags.add_argument('-dt', '--time', type=int, help='query only structures added before this time in days')
    query_flags.add_argument('-avail', '--available_values', type=str, help='list all values of field in query results')
    query_flags.add_argument('--use_source', default=False, action='store_true',
                             help='show the source rather than database ID')
    query_flags.add_argument('--since', action='store_true',
                             help='query only structures added after time specified by --time in days')
    query_flags.add_argument('--source', action='store_true',
                             help='print filenames from which structures were wrangled')
    query_flags.add_argument('-v', '--view', action='store_true',
                             help='quickly view a structure/structures with ase-gui')
    query_flags.add_argument('--cell', action='store_true',
                             help='export query to .cell files in folder name from query string')
    query_flags.add_argument('--param', action='store_true',
                             help='export query to .param files in folder name from query string')
    query_flags.add_argument('--res', action='store_true',
                             help='export query to .res files in folder name from query string')
    query_flags.add_argument('--json', action='store_true',
                             help='export query to raw json files in folder name from query string')
    query_flags.add_argument('--pdb', action='store_true',
                             help='export query to .pdb files in folder name from query string')
    query_flags.add_argument('--xsf', action='store_true',
                             help='export query to .xsf files in folder name from query string')
    query_flags.add_argument('--markdown', action='store_true', help='export query summary to a markdown file')
    query_flags.add_argument('--latex', action='store_true', help='export query summary to a LaTeX table')
    query_flags.add_argument('--write_n', type=int, help='export only those structures with n species')

    swap_flags = argparse.ArgumentParser(add_help=False)
    swap_flags.add_argument('-sw', '--swap', type=str, nargs='+',
                            help='swap all atoms in structures from a query from the first n-1 species to the nth, '
                                 'e.g. -sw NAs will swap all N to As, -sw NAs:LiNa will swap all N to As, and all Li '
                                 'to Na, and -sw [V]As:[Li,K,Rb]Na will swap all group V elements to As and all of Li,'
                                 'K and Rb to Na.')
    diff_flags = argparse.ArgumentParser(add_help=False)
    diff_flags.add_argument('-cmp', '--compare', type=str, nargs='+',
                            help='diff phase diagrams between two different times, in standard time format, '
                                 'e.g. `--compare 1y2m5d3h` will compare the present hull with that of 1 year, 2 '
                                 'months, 5 days and 3 hours ago, and `--compare 3d 2d` will compare three days ago '
                                 'to two days ago.')

    refine_flags = argparse.ArgumentParser(add_help=False)
    refine_flags.add_argument('-task', '--task', type=str,
                              help=('refine subtask to perform: options are spg, elem_set, tag, doi, source, '
                                    'pspot or raw or sub'))
    refine_flags.add_argument('-mode', '--mode', type=str,
                              help='mode of refinement: options are display, set and overwrite')
    refine_flags.add_argument('-symprec', '--symprec', type=float, help='spglib symmetry precision for refinement')
    refine_flags.add_argument('--new_tag', type=str, help='new tag to add to structures in query')
    refine_flags.add_argument('--new_doi', type=str, help='new doi to add to structures in query')

    stats_flags = argparse.ArgumentParser(add_help=False)
    stats_flags.add_argument('-l', '--list', action='store_true', help='list all collections, their sizes, and owners')
    stats_flags.add_argument('--delete', action='store_true', help='try to delete collection specified by --db')

    # define subcommand parsers and their arguments

    # matador stats
    subparsers.add_parser('stats', help='print some stats about the database.', parents=[global_flags, stats_flags])

    # matador query
    subparsers.add_parser('query',
                          help='query and extract structures from the database',
                          parents=[global_flags, query_flags, structure_flags])

    # matador import
    subparsers.add_parser('import',
                          help='import new structures in folder into database',
                          parents=[global_flags, import_flags])

    # matador hull
    subparsers.add_parser('hull',
                          help='create a convex hull from query results (currently limited to binaries and ternaries)',
                          parents=[global_flags, structure_flags,
                                   material_flags, plot_flags, query_flags])

    # matador voltage
    subparsers.add_parser('voltage',
                          help='plot a voltage curve from query results (currently limited to binaries and ternaries)',
                          parents=[global_flags, structure_flags,
                                   material_flags, plot_flags, query_flags])

    # matador changes
    subparsers.add_parser('changes',
                          help='view database changelog or undo additions to database (NB: not deletions!)',
                          parents=[global_flags, changes_flags])

    # matador hulldiff
    subparsers.add_parser('hulldiff',
                          help='diff two convex hulls with the --compare flag.',
                          parents=[global_flags, structure_flags,
                                   material_flags, plot_flags, query_flags, diff_flags])

    # matador swaps
    subparsers.add_parser('swaps',
                          help='perform atomic swaps on query results',
                          parents=[global_flags, collection_flags, query_flags,
                                   structure_flags, material_flags, swap_flags])

    # matador refine
    subparsers.add_parser('refine',
                          help='update structures in the database according to specified --task',
                          parents=[global_flags, query_flags, structure_flags,
                                   refine_flags, material_flags])

    parsed_args = parser.parse_args()
    vars_args = vars(parsed_args)

    # check for inconsistent argument combinations
    if vars_args.get('intersection') and vars_args.get('composition') is None:
        raise SystemExit('--intersection requires --composition.')
    if vars_args.get('subcmd') == 'stats' and vars_args.get('list') and vars_args.get(
            'delete'):
        raise SystemExit('Cannot use -l/--list and --delete')
    if vars_args.get('field') and vars_args.get('filter') is None:
        raise SystemExit('--field requires --filter.')
    if vars_args.get('subcmd') == 'hull' and vars_args.get('composition') is None:
        raise SystemExit('hull requires --composition')
    if vars_args.get('calc_match') and vars_args.get('id') is None:
        raise SystemExit('calc_match requires specification of a text_id with -i, exiting...')
    if vars_args.get('profile'):
        import cProfile
        import pstats
        from sys import version_info
        profiler = cProfile.Profile()
        profiler.enable()

    MatadorCommandLine(parsed_args, argstr=argv[1:], no_quickstart=no_quickstart)

    if vars_args.get('profile'):
        profiler.disable()
        fname = 'matador-{}-{}.{}.{}'.format(__version__, version_info.major, version_info.minor, version_info.micro)
        profiler.dump_stats(fname + '.prof')
        with open(fname + '.pstats', 'w') as fp:
            stats = pstats.Stats(profiler, stream=fp).sort_stats('cumulative')
            stats.print_stats()


if __name__ == '__main__':
    main()



# coding: utf-8
# Distributed under the terms of the MIT license.

""" This submodule module implements some useful exception types,
mostly for use in the :mod:`compute` and :mod:`calculator` submodules.

"""


class CalculationError(Exception):
    """ Raised when a particular calculation fails, for non-fatal reasons. """


class MaxMemoryEstimateExceeded(Exception):
    """ Raised when a structure is estimated to exceed the max memory. """


class CriticalError(RuntimeError):
    """ Raise this when you don't want any more jobs to run because something
    uncorrectable has happened! Plays more nicely with multiprocessing than
    SystemExit.

    """


class InputError(RuntimeError):
    """ Raise this when there is an issue with the input files. """


class WalltimeError(RuntimeError):
    """ Raise this when you don't want any more jobs to run
    because they're about to exceed the max walltime.

    """


class NodeCollisionError(CalculationError):
    """ Dummy exception to raise when one node has tried to run
    a calculation that another node is performing.

    """



#!/usr/bin/env python
# standard library
import unittest
import copy
from os.path import realpath

import numpy as np

# matador modules
from matador.crystal.crystal import Crystal, UnitCell
from matador.crystal.crystal_site import Site
from matador.scrapers.castep_scrapers import castep2dict, res2dict
from matador.utils.cell_utils import frac2cart
from matador.scrapers.magres_scrapers import magres2dict

# grab abs path for accessing test data
REAL_PATH = "/".join(realpath(__file__).split("/")[:-1]) + "/"

try:
    import networkx  # noqa

    imported_networkx = True
except ImportError:
    imported_networkx = False

imported_vornet = False


class UnitCellTest(unittest.TestCase):
    def test_cart_init(self):
        lattice_cart = [[3, 0, 0], [0, 3, 0], [0, 0, 3]]
        lat_tup = tuple(tuple(vec) for vec in lattice_cart)
        cell = UnitCell(lattice_cart)
        self.assertEqual(cell.lattice_cart, lat_tup)
        self.assertEqual(cell.lattice_abc, ((3, 3, 3), (90, 90, 90)))
        self.assertEqual(cell.volume, 27)
        self.assertEqual(cell.lengths, (3, 3, 3))
        self.assertEqual(cell.angles, (90, 90, 90))

        lattice_cart = np.asarray([[3, 0, 0], [0, 3, 0], [0, 0, 3]])
        cell_2 = UnitCell(lattice_cart)
        self.assertAlmostEqual(cell_2.lattice_cart, lat_tup)
        self.assertAlmostEqual(cell_2.lattice_abc, ((3, 3, 3), (90, 90, 90)))
        self.assertEqual(cell_2.volume, 27)
        self.assertAlmostEqual(cell_2.lengths, (3, 3, 3))
        self.assertAlmostEqual(cell_2.angles, (90, 90, 90))
        self.assertEqual(cell.lattice_cart, lat_tup)
        self.assertEqual(cell.lattice_abc, ((3, 3, 3), (90, 90, 90)))
        self.assertEqual(cell.volume, 27)
        self.assertEqual(cell.lengths, (3, 3, 3))

        lattice_cart = [[10, 0, 0], [0, 10, 0], [0, 0, 10]]
        cell.lattice_cart = lattice_cart
        lat_tup = tuple(tuple(vec) for vec in lattice_cart)
        self.assertEqual(cell.lattice_cart, lat_tup)
        lattice_cart = "aadsfadsf"
        self.assertEqual(cell.lattice_cart, lat_tup)
        self.assertEqual(cell.lattice_abc, ((10, 10, 10), (90, 90, 90)))
        self.assertEqual(cell.volume, 1000)

    def test_abc_init(self):
        lattice_abc = [[2, 3, 4], [60, 60, 60]]
        lat_tup = tuple(tuple(elem) for elem in lattice_abc)
        cell = UnitCell(lattice_abc)
        self.assertAlmostEqual(cell.lattice_abc, lat_tup)
        cell.lengths = [10, 10, 10]
        self.assertEqual(cell.lattice_abc, ((10, 10, 10), (60, 60, 60)))
        cell.angles = [90, 90, 90]
        self.assertEqual(cell.lattice_abc, ((10, 10, 10), (90, 90, 90)))
        lattice_cart = ((10, 0, 0), (0, 10, 0), (0, 0, 10))
        self.assertEqual(cell.lattice_cart, lattice_cart)


class CrystalTest(unittest.TestCase):
    def test_getters_setters(self):
        doc, s = castep2dict(REAL_PATH + "data/Na3Zn4-swap-ReOs-OQMD_759599.castep")
        crystal = Crystal(doc)
        self.assertEqual(
            list(crystal.lattice_cart[0]), [9.0397727, 0.0081202, 0.0000000]
        )
        self.assertEqual(crystal.num_atoms, 14)
        with self.assertRaises(AttributeError):
            crystal["positions_frac"] = [[0, 1, 2]]

        # check we can set fields to the same value
        crystal["new_field"] = [1, 2, 3]
        crystal["new_field"] = [1, 2, 3]

        crystal["new_field_2"] = np.nan
        crystal["new_field_2"] = np.nan

        crystal["new_field_3"] = [1, 2, 4]
        with self.assertRaises(AttributeError):
            crystal["new_field_3"] = [1, 2, 5]

        crystal["new_field_4"] = [1, 2, np.nan]
        crystal["new_field_4"] = [1, 2, np.nan]

        crystal["new_field_5"] = [1, np.nan, 2]
        with self.assertRaises(AttributeError):
            crystal["new_field_5"] = [1, 2, np.nan]

        crystal["new_field_6"] = np.linspace(0, 1, 1000).tolist()
        crystal["new_field_6"] = np.array(crystal["new_field_6"], copy=True).tolist()

    def test_set_positions(self):
        doc, s = castep2dict(REAL_PATH + "data/Na3Zn4-swap-ReOs-OQMD_759599.castep")
        doc = Crystal(doc)

        copydoc = copy.deepcopy(doc)
        old_pos = np.asarray(doc.positions_frac)
        copydoc.set_positions(np.zeros_like(old_pos), fractional=True)

        np.testing.assert_array_almost_equal(
            np.asarray(copydoc.positions_frac), np.zeros_like(old_pos)
        )
        np.testing.assert_array_almost_equal(
            np.asarray(copydoc.positions_abs), np.zeros_like(old_pos)
        )

        self.assertNotAlmostEqual(doc.positions_frac[-1][0], 0.0)

    def test_convert_positions(self):
        doc = res2dict(REAL_PATH + "data/structures/Li7Sn-Fmmm.res")[0]
        crystal = res2dict(REAL_PATH + "data/structures/Li7Sn-Fmmm.res", as_model=True)[
            0
        ]

        doc["positions_abs"] = frac2cart(doc["lattice_cart"], doc["positions_frac"])

        np.testing.assert_array_equal(doc["positions_abs"], crystal.positions_abs)
        print(crystal.positions_abs)
        for site in crystal:
            print(site, site.get_coords("cartesian"))

    def test_minimal_init(self):
        doc = Crystal(
            dict(
                lattice_abc=np.asarray([[3, 3, 3], [90, 90, 90]]),
                atom_types=["Na", "Cl"],
                positions_frac=[[0, 0, 0], [0.5, 0.5, 0.5]],
            )
        )
        self.assertEqual(doc.stoichiometry, [["Cl", 1.0], ["Na", 1.0]])
        self.assertEqual(doc.lattice_abc, ((3.0, 3.0, 3.0), (90.0, 90.0, 90.0)))
        self.assertEqual(
            doc.lattice_cart, ((3.0, 0.0, 0.0), (0.0, 3.0, 0.0), (0.0, 0.0, 3.0))
        )
        self.assertEqual(len(doc.sites), 2)
        self.assertEqual(doc.num_atoms, 2)
        self.assertEqual(doc.concentration, [0.5, 0.5])
        self.assertEqual(doc.positions_abs, [[0, 0, 0], [1.5, 1.5, 1.5]])
        self.assertEqual(doc.positions_frac, [[0, 0, 0], [0.5, 0.5, 0.5]])
        self.assertEqual(doc.formula, "ClNa")
        self.assertEqual(doc.cell_volume, 27.0)
        self.assertEqual(doc.space_group, "Pm-3m")

        doc = Crystal(
            dict(
                lattice_cart=((3.0, 0.0, 0.0), (0.0, 3.0, 0.0), (0.0, 0.0, 3.0)),
                atom_types=["Na", "Cl"],
                positions_abs=[[0, 0, 0], [1.5, 1.5, 1.5]],
            )
        )
        self.assertEqual(doc.lattice_abc, ((3.0, 3.0, 3.0), (90.0, 90.0, 90.0)))
        self.assertEqual(
            doc.lattice_cart, ((3.0, 0.0, 0.0), (0.0, 3.0, 0.0), (0.0, 0.0, 3.0))
        )
        self.assertEqual(doc.stoichiometry, [["Cl", 1.0], ["Na", 1.0]])
        self.assertEqual(len(doc.sites), 2)
        self.assertEqual(doc.num_atoms, 2)
        self.assertEqual(doc.concentration, [0.5, 0.5])
        self.assertEqual(doc.positions_abs, [[0.0, 0.0, 0.0], [1.5, 1.5, 1.5]])
        self.assertEqual(doc.positions_frac, [[0, 0, 0], [0.5, 0.5, 0.5]])
        self.assertEqual(doc.formula, "ClNa")
        self.assertEqual(doc.cell_volume, 27.0)
        self.assertEqual(doc.space_group, "Pm-3m")

    def testSites(self):
        doc, s = castep2dict(REAL_PATH + "data/Na3Zn4-swap-ReOs-OQMD_759599.castep")
        del doc["lattice_cart"]
        crystal = Crystal(doc)
        print(crystal)
        self.assertEqual(crystal[0].coords, [0.776467, 0.466319, 0.0])
        with self.assertRaises(RuntimeError):
            crystal[0].set_position([0.5, 0.6, 0.7, 0.8], "fractional")
        with self.assertRaises(RuntimeError):
            crystal[0].set_position([[1, 2, 3], [4, 5, 6], [7, 8, 9]], "fractional")
        self.assertEqual(
            [atom for atom in crystal], [atom[1] for atom in enumerate(crystal)]
        )

        atom = Site(
            species="Cl",
            position=[0.2, 0.5, 0.2],
            lattice_cart=[[10, 0, 0], [0, 10, 0], [0, 0, 10]],
        )
        atom2 = copy.deepcopy(atom)
        atom2.species = "Br"

        self.assertEqual(atom.species, "Cl")
        self.assertEqual(atom2.species, "Br")

        atom2.set_position([1.2, -0.5, 0.2], "fractional")
        np.testing.assert_array_almost_equal(
            atom2.displacement_between_sites(atom), [0.0, 0.0, 0.0], decimal=10
        )
        self.assertAlmostEqual(atom2.distance_between_sites(atom), 0.0, places=10)
        atom2.set_position([1.3, -0.5, 0.2], "fractional")
        np.testing.assert_array_almost_equal(
            atom2.displacement_between_sites(atom), [1.0, 0.0, 0.0], decimal=10
        )
        self.assertAlmostEqual(atom2.distance_between_sites(atom), 1.0, places=10)
        atom2.set_position([1.3, -0.5, 0.3], "fractional")
        np.testing.assert_array_almost_equal(
            atom2.displacement_between_sites(atom), [1.0, 0.0, 1.0], decimal=10
        )
        self.assertAlmostEqual(
            atom2.distance_between_sites(atom), np.sqrt(2), places=10
        )

    def testSpg(self):
        doc, s = castep2dict(REAL_PATH + "data/Na3Zn4-swap-ReOs-OQMD_759599.castep")
        crystal = Crystal(doc)
        print(crystal.get_space_group(symprec=0.01))
        print(crystal.get_space_group(symprec=0.001))
        self.assertEqual(crystal.get_space_group(symprec=0.0000001), "Pm")

    def testFromMagres(self):
        doc, s = magres2dict(REAL_PATH + "data/NaP_QE6.magres")
        crystal = Crystal(doc)
        for atom in crystal:
            print(atom, atom.site_data["magres_shift"])

    @unittest.skipIf(not imported_vornet, "Voronoi code not found in this distribution")
    def testCoordination(self):
        doc, s = magres2dict(REAL_PATH + "data/NaP_QE6.magres")
        crystal = Crystal(doc, voronoi=True)
        for atom in crystal:
            print(atom, atom.coordination)
        print(crystal.coordination_lists)
        print(crystal.coordination_stats)

    @unittest.skipIf(not imported_vornet, "Voronoi code not found in this distribution")
    def testVoronoi(self):
        doc, s = magres2dict(REAL_PATH + "data/NaP_QE6.magres")
        crystal = Crystal(doc)
        print(crystal.unique_sites)

    @unittest.skipIf(not imported_networkx, "NetworkX missing")
    def testBondLengths(self):
        doc, s = magres2dict(REAL_PATH + "data/NaP_QE6.magres")
        crystal = Crystal(doc)
        print(crystal.bond_lengths)

    @unittest.skipIf(not imported_networkx, "NetworkX missing")
    def testBondStats(self):
        doc, s = magres2dict(REAL_PATH + "data/NaP_QE6.magres")
        crystal = Crystal(doc)
        print(crystal.bonding_stats)


class ElasticCrystalTest(unittest.TestCase):
    """ Test the elastic functionality of the Crystal module. """

    def testKBulkModulus(self):
        from matador.crystal.elastic import get_equation_of_state

        results = get_equation_of_state(
            REAL_PATH + "/data/bulk_modulus/K-bulk_modulus", plot=False
        )
        self.assertTrue("eos" in results)
        self.assertEqual(len(results["eos"]), 3)
        self.assertAlmostEqual(results["eos"][0].bulk_modulus, 3.696117355)
        self.assertAlmostEqual(results["eos"][1].bulk_modulus, 3.699072676)
        self.assertAlmostEqual(results["eos"][2].bulk_modulus, 3.691406442)
        self.assertAlmostEqual(results["eos"][0].bulk_modulus_err, 3e-6, places=1)
        self.assertAlmostEqual(results["eos"][1].bulk_modulus_err, 2e-6, places=1)
        self.assertAlmostEqual(results["eos"][2].bulk_modulus_err, 2e-6, places=1)


if __name__ == "__main__":
    unittest.main(buffer=False, verbosity=2)



""" This submodule includes some useful utility functions
for exporting data.

"""

from traceback import print_exc
import os
import random


def file_writer_function(function):
    """ Wrapper for file writers to safely overwrite/hash duplicate files.

    Keyword arguments:
        overwrite (bool): whether or not to overwrite colliding files.
        hash_dupe (bool): whether or not to create a unique filename for
            any colliding files, or just skip writing them.

    """

    from functools import wraps

    @wraps(function)
    def wrapped_writer(doc, path, overwrite=False, hash_dupe=False, **kwargs):
        """ Wrap and return the writer function. """
        try:
            overwrite = kwargs.pop('overwrite', overwrite)
            hash_dupe = kwargs.pop('hash_dupe', hash_dupe)
            flines, ext = function(doc, path, overwrite=overwrite, hash_dupe=hash_dupe, **kwargs)
            if ext is not None and not path.endswith('.' + ext):
                path += '.{}'.format(ext)

            if os.path.isfile(path):
                if overwrite:
                    os.remove(path)
                elif hash_dupe:
                    print('File name already exists, generating hash...')
                    path = '{}-{}.{}'.format(path.replace(ext, ''), generate_hash(), ext)
                else:
                    print('File name {} already exists! Skipping!'.format(path))
                    return False

            with open(path, 'w') as f:
                for line in flines:
                    f.write(line + '\n')

        except Exception as exc:
            print_exc()
            raise type(exc)('Failed to write {}: {}'.format(path, exc))

    return wrapped_writer


def generate_hash(hash_len=6):
    """ Quick hash generator, based on implementation in PyAIRSS by J. Wynn.

    Keyword arguments:
        hash_len (int): desired length of hash.

    """
    hash_chars = [str(x) for x in range(10)] + [chr(97+i) for i in range(25)]
    _hash = ''
    for _ in range(hash_len):
        _hash += random.choice(hash_chars)
    return _hash


def generate_relevant_path(**args):
    """ Generates a suitable path name based on query. """
    dirname = ''
    if args.get('subcmd') is not None:
        dirname += args.get('subcmd') + '-'
    else:
        dirname = 'query'
    if args.get('composition') is not None:
        for comp in args['composition']:
            dirname += comp
    elif args.get('formula') is not None:
        dirname += args.get('formula')[0]
    if args.get('db') is not None:
        dirname += '-' + args.get('db')[0]
    if args.get('swap') is not None:
        for swap in args['swap']:
            dirname += '-' + swap
        if args.get('hull_cutoff') is not None:
            dirname += '-hull-' + str(args.get('hull_cutoff')) + 'eV'
    if args.get('id') is not None:
        dirname += '-' + args.get('id')[0] + '_' + args.get('id')[1]
    dirname = dirname.replace('--', '-')
    return dirname



# coding: utf-8
""" This file implements some prototype scrapers of
experimental .synth and .expt files.
"""

from __future__ import print_function
# external libraries
import bson.json_util as json
# standard library
from collections import defaultdict


def synth2dict(seed, **kwargs):
    """ Take a .synth file and create a dict
    suitable for database insertion.
    """
    synth = defaultdict(list)
    if seed.endswith('.synth'):
        seed = seed.replace('.synth', '')
    with open(seed+'.synth', 'r') as f:
        flines = f.readlines()
    synth['source'].append(seed+'.synth')
    # define splitters
    splitters = [':', '=']
    try:
        for line_no, line in enumerate(flines):
            # skip blank lines and comments
            if line.startswith(('#', '!')) or len(line.strip()) == 0:
                continue
            else:
                # check if line is split further by semi-colons
                if ';' in line:
                    line = line.split(';')
                    for ind, entry in enumerate(line):
                        line[ind] = line[ind].strip()
                # else set all lines to be single item lists so that
                # we can scrape lines and ;-terminated lines in the same way
                else:
                    line = [line]
                # now look for splitters between tags and data
                for entry in line:
                    for splitter in splitters:
                        if splitter in entry:
                            synth[entry.split(splitter)[0].strip()] = entry.split(splitter)[-1].replace('\n', '')
    except Exception as oopsy:
        if kwargs.get('verbosity') > 0:
            print(oopsy)
            print('Error in', seed+'.synth, skipping...')
        return seed + '\t\t' + str(oopsy), False
    if kwargs.get('debug'):
        print(json.dumps(synth, indent=2, ensure_ascii=False))
    return synth, True


def expt2dict(seed, **kwargs):
    """ Take a .expt file and create a dict
    suitable for database insertion.
    """
    expt = defaultdict(list)
    if seed.endswith('.expt'):
        seed = seed.replace('.expt', '')
    with open(seed+'.expt', 'r') as f:
        flines = f.readlines()
    expt['source'].append(seed+'.expt')
    # define splitters
    splitters = [':', '=']
    try:
        for line_no, line in enumerate(flines):
            # skip blank lines and comments
            if line.startswith(('#', '!')) or len(line.strip()) == 0:
                continue
            else:
                # check if line is split further by semi-colons
                if ';' in line:
                    line = line.split(';')
                    for ind, entry in enumerate(line):
                        line[ind] = line[ind].strip()
                # else set all lines to be single item lists so that
                # we can scrape lines and ;-terminated lines in the same way
                else:
                    line = [line]
                # now look for splitters between tags and data
                for entry in line:
                    for splitter in splitters:
                        if splitter in entry:
                            if entry.split(splitter)[0].strip() in expt:
                                expt[entry.split(splitter)[0].strip()] = [expt[entry.split(splitter)[0].strip()]]
                                expt[entry.split(splitter)[0].strip()].append(entry.split(splitter)[-1].replace('\n', ''))
                            else:
                                expt[entry.split(splitter)[0].strip()] = entry.split(splitter)[-1].replace('\n', '')
    except Exception as oopsy:
        if kwargs.get('verbosity') > 0:
            print(oopsy)
            print('Error in', seed+'.expt, skipping...')
        return seed + '\t\t' + str(oopsy), False
    if kwargs.get('debug'):
        print(json.dumps(expt, indent=2, ensure_ascii=False))
    return expt, True



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule contains a dirty wrapper of ase-gui for quick
visualisation, and nglview wrapper for JupyterNotebook visualisation,
and some colour definitions scraped from VESTA configs.

"""

from matador.utils.ase_utils import doc2ase


def viz(doc):
    """ Quick and dirty ase-gui visualisation from matador doc. """
    from ase.visualize import view
    view(doc2ase(doc))
    return


def get_element_colours():
    """ Read element colours from VESTA file. The colours file can be
    specified in the matadorrc. If unspecified, the default
    ../config/vesta_elements.ini will be used.

    """
    import os
    from matador.config import load_custom_settings, SETTINGS
    # check if element_colours has been given as an absolute path
    if SETTINGS:
        colours_fname = SETTINGS.get('plotting', {}).get('element_colours')
    else:
        colours_fname = load_custom_settings().get('plotting', {}).get('element_colours')

    if colours_fname is None:
        colours_fname = '/'.join(os.path.realpath(__file__).split('/')[:-1]) + '/../config/vesta_elements.ini'
    elif not os.path.isfile(colours_fname):
        print('Could not find {}, please specify an absolute path. Falling back to default...'.format(colours_fname))
        # otherwise fallback to ../config/vesta_elements.ini
        colours_fname = '/'.join(os.path.realpath(__file__).split('/')[:-1]) + '/../config/vesta_elements.ini'

    with open(colours_fname, 'r') as f:
        flines = f.readlines()
    element_colours = dict()
    for line in flines:
        line = line.split()
        elem = line[1]
        colour = list(map(float, line[-3:]))
        element_colours[elem] = colour
    return element_colours


def nb_viz(doc, repeat=1, bonds=None):
    """ Return an ipywidget for nglview visualisation in
    a Jupyter notebook or otherwise.

    Parameters:
        doc (matador.crystal.Crystal / dict): matador document to show.

    Keyword arguments:
        repeat (int): number of periodic images to include.
        bonds (str): custom bond selection.

    """
    import nglview
    atoms = doc2ase(doc)
    atoms = atoms.repeat((repeat, repeat, repeat))
    view = nglview.show_ase(atoms)
    view.add_unitcell()
    view.remove_ball_and_stick()
    view.add_spacefill(radius_type='vdw', scale=0.3)
    if bonds is None:
        for elements in set(doc['atom_types']):
            view.add_ball_and_stick(selection='#{}'.format(elements.upper()))
    else:
        if not isinstance(bonds, list):
            bonds = [bonds]
        for bond in bonds:
            view.add_ball_and_stick(selection='#{}'.format(bond.upper()))
    view.parameters = {'clipDist': 0}
    view.camera = 'orthographic'
    view.background = '#FFFFFF'
    view.center()
    return view



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements some scraper functions for
Quantum Espresso-related inputs and outputs.

"""


from collections import defaultdict
from os import stat
from math import gcd
from pwd import getpwuid
from matador.utils.cell_utils import cart2abc, cart2volume
from matador.utils.chem_utils import RY_TO_EV, KBAR_TO_GPA
from matador.scrapers.utils import scraper_function


@scraper_function
def pwout2dict(seed, **kwargs):
    """ Extract available information from pw.x .out file.

    Parameters:
        seed (str/list): filename or list of filenames to scrape as a
            QuantumEspresso pw.x output.

    """
    pwout = defaultdict(list)
    # read .pwout file into array
    if seed.endswith('.out'):
        ext = '.out'
        seed = seed.replace('.out', '')
    elif seed.endswith('.in'):
        ext = '.in'
        seed = seed.replace('.in', '')
    else:
        ext = seed.split('.')[-1]

    with open(seed+ext, 'r') as f:
        flines = f.readlines()
    # add .pwout to source
    pwout['source'].append(seed+ext)
    # grab file owner username
    pwout['user'] = getpwuid(stat(seed+ext).st_uid).pw_name
    if 'CollCode' in seed:
        pwout['icsd'] = seed.split('CollCode')[-1]
    for ind, line in enumerate(reversed(flines)):
        ind = len(flines) - 1 - ind
        if 'cell_parameters' in line.lower() and 'angstrom' in line.lower() and 'lattice_cart' not in pwout:
            pwout['lattice_cart'] = []
            for j in range(3):
                line = flines[ind+j+1].strip().split()
                pwout['lattice_cart'].append(list(map(float, line)))
            pwout['cell_volume'] = cart2volume(pwout['lattice_cart'])
        elif 'atomic_positions' in line.lower() and 'positions_frac' not in pwout:
            pwout['positions_frac'] = []
            pwout['atom_types'] = []
            j = 1
            while True:
                if 'End final coordinates' in flines[j+ind]:
                    break
                else:
                    try:
                        line = flines[j+ind].strip().split()
                        pwout['atom_types'].append(line[0])
                        pwout['positions_frac'].append(list(map(float, line[1:5])))
                        j += 1
                    except Exception:
                        break
            pwout['num_atoms'] = len(pwout['atom_types'])
        elif 'final enthalpy' in line.lower() and 'enthalpy' not in pwout:
            pwout['enthalpy'] = RY_TO_EV * float(line.lower().split()[-2])
        elif 'total   stress' in line.lower() and 'pressure' not in pwout:
            pwout['pressure'] = KBAR_TO_GPA * float(line.lower().split()[-1])
        elif all(key in pwout for key in ['enthalpy', 'pressure', 'lattice_cart', 'positions_frac']):
            break
    # get abc lattice
    pwout['lattice_abc'] = cart2abc(pwout['lattice_cart'])
    # calculate stoichiometry
    pwout['stoichiometry'] = defaultdict(float)
    for atom in pwout['atom_types']:
        if atom not in pwout['stoichiometry']:
            pwout['stoichiometry'][atom] = 0
        pwout['stoichiometry'][atom] += 1
    gcd_val = 0
    for atom in pwout['atom_types']:
        if gcd_val == 0:
            gcd_val = pwout['stoichiometry'][atom]
        else:
            gcd_val = gcd(pwout['stoichiometry'][atom], gcd_val)
    # convert stoichiometry to tuple for fryan
    temp_stoich = []
    for key, value in pwout['stoichiometry'].items():
        if float(value)/gcd_val % 1 != 0:
            temp_stoich.append([key, float(value)/gcd_val])
        else:
            temp_stoich.append([key, value/gcd_val])
    pwout['stoichiometry'] = temp_stoich
    atoms_per_fu = 0
    for elem in pwout['stoichiometry']:
        atoms_per_fu += elem[1]
    pwout['num_fu'] = len(pwout['atom_types']) / atoms_per_fu

    return pwout, True



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The db module provides all the submodules that touch the database,
with functionality to connect, add or refine database objects, and observe
changes.

"""


__all__ = ['Spatula', 'DatabaseChanges', 'Refiner', 'make_connection_to_collection']
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'

from matador.db.connect import make_connection_to_collection
from matador.db.importer import Spatula
from matador.db.changes import DatabaseChanges
from matador.db.refine import Refiner
